#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-21
# @Filename: test_sky.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import sys

import astropy.time
import numpy
import pytest
from astropy import units as u
from astropy.coordinates import AltAz, Distance, EarthLocation, SkyCoord
from coordio import ICRS, CoordIOError, Observed, Site

wavelength = 7000

icrs = ICRS([[100, 10], [101., 11]],
            pmra=[100, 0.1], pmdec=[-15, -0.5],
            parallax=[10., 10.], rvel=[100, 1000],
            epoch=[2451545, 2451545])


astropy_icrs = SkyCoord(ra=[100, 101] * u.deg, dec=[10, 11] * u.deg,
                        frame='icrs',
                        pm_ra_cosdec=[100, 0.1] * u.mas / u.yr,
                        pm_dec=[-15., -0.5] * u.mas / u.yr,
                        distance=Distance(parallax=[10., 10.] * u.mas),
                        radial_velocity=[100, 1000] * u.km / u.s,
                        obstime=astropy.time.Time(2451545.0,
                                                  format='jd',
                                                  scale='tai'))

site = Site('APO')
astropy_location = EarthLocation.from_geodetic(lon=site.longitude,
                                               lat=site.latitude,
                                               height=site.altitude)


def test_icrs():

    icrs = ICRS([[100, 10], [101., 11]])

    assert isinstance(icrs, ICRS)
    assert (icrs.epoch == 2451545.0).all()


def test_to_epoch():

    icrs_2020 = icrs.to_epoch(jd=2458863, site=site)

    new_obstime = astropy.time.Time(2458863, format='jd', scale='tdb')
    new_astropy_icrs = astropy_icrs.apply_space_motion(new_obstime=new_obstime)

    numpy.testing.assert_array_almost_equal(new_astropy_icrs.ra.deg,
                                            icrs_2020[:, 0],
                                            decimal=6)
    numpy.testing.assert_array_almost_equal(new_astropy_icrs.dec.deg,
                                            icrs_2020[:, 1],
                                            decimal=6)

    if sys.version_info < (3, 8):
        cos_dec = numpy.cos(numpy.radians(new_astropy_icrs.dec.deg))
        astropy_pm_ra_cosdec = new_astropy_icrs.pm_ra.value * cos_dec
    else:
        astropy_pm_ra_cosdec = new_astropy_icrs.pm_ra_cosdec.value

    numpy.testing.assert_array_almost_equal(astropy_pm_ra_cosdec,
                                            icrs_2020.pmra,
                                            decimal=6)
    numpy.testing.assert_array_almost_equal(new_astropy_icrs.pm_dec.value,
                                            icrs_2020.pmdec,
                                            decimal=6)


def test_to_observed():

    site.set_time(2458863, scale='TAI')

    # test for failure if site isn't provided
    with pytest.raises(CoordIOError):
        observed = Observed(icrs)

    observed = Observed(icrs, site=site, wavelength=wavelength)

    new_obstime = astropy.time.Time(2458863, format='jd', scale='tai')
    new_astropy_icrs = astropy_icrs.apply_space_motion(new_obstime=new_obstime)
    astropy_observed = new_astropy_icrs.transform_to(
        AltAz(location=astropy_location,
              obstime=new_obstime,
              pressure=site.pressure * u.mbar,
              temperature=site.temperature * u.deg_C,
              relative_humidity=site.rh,
              obswl=wavelength * u.Angstrom))

    assert isinstance(observed, Observed)

    numpy.testing.assert_allclose(observed[:, 0],
                                  astropy_observed.alt.deg,
                                  atol=3e-5, rtol=1e-7)
    numpy.testing.assert_allclose(
        observed[:, 1], astropy_observed.az.deg,
        atol=3e-5 / numpy.cos(numpy.radians(astropy_observed.alt.deg)).max(),
        rtol=1e-7)

    # test that raw initialization of observed yeilds the same answers
    obsArr = numpy.array(observed)
    observed2 = Observed(obsArr, site=site)

    numpy.testing.assert_allclose(observed[:, 0], observed2[:, 0],
                                  atol=3e-7, rtol=0)

    numpy.testing.assert_allclose(observed[:, 1], observed2[:, 1],
                                  atol=3e-7, rtol=0)

    numpy.testing.assert_allclose(observed.ra, observed2.ra,
                                  atol=3e-7, rtol=0)

    numpy.testing.assert_allclose(observed.dec, observed2.dec,
                                  atol=3e-7, rtol=0)

    numpy.testing.assert_allclose(observed.ha, observed2.ha,
                                  atol=3e-7, rtol=0)

    numpy.testing.assert_allclose(observed.pa, observed2.pa,
                                  atol=3e-7, rtol=0)

    # _icrs = ICRS(observed, wavelength=wavelength)
    # numpy.testing.assert_array_almost_equal(icrs, _icrs, decimal=5)


def test_icrs_obs_cycle():
    time = 2451545
    icrs = ICRS([[100, 10], [101., 11], [180., 30]], epoch=time)

    site.set_time(time, scale='TAI')
    observed = Observed(icrs, site=site)
    _icrs = ICRS(observed)

    a1 = SkyCoord(ra=icrs[:, 0] * u.deg, dec=icrs[:, 1] * u.deg)
    a2 = SkyCoord(ra=_icrs[:, 0] * u.deg, dec=_icrs[:, 1] * u.deg)
    sep = a1.separation(a2)
    assert numpy.max(numpy.array(sep) * 3600) < 0.5


@pytest.mark.skipif(sys.version_info < (3, 8), reason='requires python 3.8+ for HADec')
@pytest.mark.parametrize('hadec', [True, False])
def test_observed_from_equatorial(hadec):
    from astropy.coordinates import HADec

    time = 2451545
    site.set_time(time, scale='TAI')

    ras = numpy.random.uniform(0, 360, 100)
    decs = numpy.random.uniform(-90, 90, 100)

    # Get HA for cases when hadec=True
    astropy_time = astropy.time.Time(time, format='jd', scale='tai',
                                     location=astropy_location)
    astropy_lst = astropy_time.sidereal_time('mean').deg

    ha_d = astropy_lst - ras
    if hadec:
        eqs = numpy.array((ha_d, decs)).T
    else:
        eqs = numpy.array([ras, decs]).T

    observed = Observed.fromEquatorial(eqs, hadec=hadec, site=site,
                                       wavelength=wavelength)

    assert observed.shape == (100, 2)
    assert numpy.all(observed[:, 0] > -90) and numpy.all(observed[:, 0] < 90)
    assert numpy.all(observed[:, 1] > 0) and numpy.all(observed[:, 1] < 360)

    # Now calculate the same using astropy and compare.
    hadec_coords = HADec(ha=ha_d * u.deg, dec=decs * u.deg,
                         obstime=astropy_time, location=astropy_location)

    astropy_altaz = hadec_coords.transform_to(AltAz(location=astropy_location,
                                                    obstime=astropy_time))

    assert numpy.allclose(observed[:, 0], astropy_altaz.alt.deg, atol=1e-6)
    assert numpy.allclose(observed[:, 1], astropy_altaz.az.deg % 360, atol=1e-6)



from coordio.utils import radec2wokxy, wokxy2radec
import time
import matplotlib.pyplot as plt
import numpy
import coordio.fitData as fitData
import os


from astropy.coordinates import SkyCoord
from astropy import units as u

filedir = os.path.dirname(os.path.abspath(__file__))

# apo plate 15017
apo = {}
apo["utcJD"] = 2459249.6184
apo["alt"] = 54  # at the JD supplied...
apo["file"] = os.path.join(filedir, "plPlugMapP-15017.par")

# lco plate 12377
lco = {}
lco["utcJD"] = 2459249.8428
lco["alt"] = 45.18  # at the JD supplied
lco["file"] = os.path.join(filedir, "plPlugMapP-12377.par")


def parsePlugmap(plPlugFile):
    with open(plPlugFile) as f:
        rawLines = f.readlines()

    info = {}
    info["xFocal"] = []
    info["yFocal"] = []
    info["fiberType"] = []
    info["ra"] = []
    info["dec"] = []

    for line in rawLines:
        if line.startswith("ha "):
            info["ha"] = float(line.split()[1])
        if line.startswith("temp "):
            info["temp"] = float(line.split()[-1])
        if line.startswith("raCen "):
            info["raCen"] = float(line.split()[-1])
        if line.startswith("decCen "):
            info["decCen"] = float(line.split()[-1])
        if line.startswith("PLUGMAPOBJ "):
            if " QSO " in line:
                # assume this is a boss fiber
                split = line.split(" QSO ")[-1].split()
                info["xFocal"].append(float(split[0]))
                info["yFocal"].append(float(split[1]))

                split = line.split(" OBJECT ")[-1].split()
                info["ra"].append(float(split[0]))
                info["dec"].append(float(split[1]))

                info["fiberType"].append("Boss")
            elif " STAR_BHB " in line:
                split = line.split(" STAR_BHB ")[-1].split()
                info["xFocal"].append(float(split[0]))
                info["yFocal"].append(float(split[1]))

                split = line.split(" OBJECT ")[-1].split()
                info["ra"].append(float(split[0]))
                info["dec"].append(float(split[1]))

                info["fiberType"].append("Apogee")

    return info


def test_scales(plot=False):
    for siteName, data in zip(["LCO", "APO"], [lco,apo]):
        plateData = parsePlugmap(data["file"])

        xWok, yWok, fieldWarn, ha, pa = radec2wokxy(
            plateData["ra"], plateData["dec"], data["utcJD"], plateData["fiberType"],
            plateData["raCen"], plateData["decCen"], 0, siteName, data["utcJD"]
        )

        xWok2, yWok2, fieldWarn2, ha2, pa2 = radec2wokxy(
            plateData["ra"], plateData["dec"], data["utcJD"], plateData["fiberType"],
            plateData["raCen"], plateData["decCen"], 0, siteName, data["utcJD"],
            focalScale=1
        )

        dx = xWok2 - xWok
        dy = yWok2 - yWok

        if plot:
            plt.figure()
            plt.quiver(xWok,yWok,dx,dy,angles='xy',units='xy', scale=0.01)
            plt.axis("equal")
            plt.title(siteName)

        ra, dec, fieldWarn = wokxy2radec(xWok,yWok,plateData["fiberType"],
            plateData["raCen"], plateData["decCen"], 0, siteName, data["utcJD"])

        ra2, dec2, fieldWarn2 = wokxy2radec(xWok2,yWok2,plateData["fiberType"],
            plateData["raCen"], plateData["decCen"], 0, siteName, data["utcJD"],
            focalScale=1)

        dra = plateData["ra"] - ra
        ddec = plateData["dec"] - dec
        err = numpy.sqrt(dra**2+ddec**2)

        dra2 = plateData["ra"] - ra2
        ddec2 = plateData["dec"] - dec2
        err2 = numpy.sqrt(dra2**2+ddec2**2)

        if plot:
            plt.figure()
            plt.title(siteName + " 1")
            plt.hist(err)

            plt.figure()
            plt.title(siteName + " 2")
            plt.hist(err2)

        assert numpy.max(err) < 2e-6
        assert numpy.max(err2) < 2e-6

    # plt.show()

        # import pdb; pdb.set_trace()


def run_field(siteName, plot=False):

    if siteName == "LCO":
        dd = lco
    else:
        dd = apo
    plateData = parsePlugmap(dd["file"])

    xWok, yWok, fieldWarn, ha, pa = radec2wokxy(
        plateData["ra"], plateData["dec"], dd["utcJD"], plateData["fiberType"],
        plateData["raCen"], plateData["decCen"], 0, siteName, dd["utcJD"]
    )

    dHA = ha - plateData["ha"]

    # convert to hours
    dHA = 24/360.*dHA

    # convert to days
    dHA = dHA/24

    # update time of observation to be at designed hour angle
    timeObs = dd["utcJD"] - dHA

    xWok, yWok, fieldWarn, ha, pa = radec2wokxy(
        plateData["ra"], plateData["dec"], timeObs, plateData["fiberType"],
        plateData["raCen"], plateData["decCen"], 0, siteName, timeObs
    )

    print("obs ha", ha)
    print("design ha", plateData["ha"])
    xFocal = numpy.array(plateData["xFocal"])
    yFocal = numpy.array(plateData["yFocal"])

    # lco xy is backwards
    if siteName == "LCO":
        xFocal = xFocal*-1
        yFocal = yFocal*-1

    # if plot:
    #     plt.figure(figsize=(8,8))
    #     plt.plot(xFocal, yFocal, 'x')
    #     plt.axis("equal")
    #     plt.title("focal")

    #     plt.figure(figsize=(8,8))
    #     plt.plot(xWok, yWok, 'x')
    #     plt.axis("equal")
    #     plt.title("wok")

    dx = xFocal - xWok
    dy = yFocal - yWok
    rmsErr = numpy.sqrt(numpy.mean(dx**2+dy**2))*1000

    if plot:

        plt.figure()
        plt.title("%s\nRaw RMS error %.2f microns"%(siteName, rmsErr))
        plt.hist(numpy.sqrt(dx**2+dy**2)*1000)
        plt.xlabel("err (micron)")

        plt.figure(figsize=(8,8))
        plt.title(
            "%s\nRaw Residuals\nRMS error %.2f microns"%(siteName, rmsErr)
        )
        plt.quiver(xWok,yWok,dx,dy, angles="xy")
        plt.xlabel("x wok (mm)")
        plt.ylabel("y wok (mm)")
        plt.axis("equal")
        plt.savefig("%sErrs.png"%siteName, dpi=150)

    # fit translation, rotation, scale
    fitTransRotScale = fitData.ModelFit(
        model=fitData.TransRotScaleModel(),
        measPos=numpy.array([xWok, yWok]).T,
        nomPos=numpy.array([xFocal, yFocal]).T,
        doRaise=True,
    )

    xyOff, rotAngle, scale = fitTransRotScale.model.getTransRotScale()
    print("xy translation (micron)", xyOff * 1000)
    print("rot (deg)", rotAngle)
    print("scale", scale)

    posErr = fitTransRotScale.getPosError()
    print("posErr shape", posErr.shape)
    rmsErr = numpy.sqrt(numpy.mean(posErr[:,0]**2 + posErr[:,1]**2))*1000
    print("fit rms error (micron)", rmsErr)

    if siteName == "LCO":
        assert rmsErr < 1

    if plot:
        # plt.figure()
        # plt.hist(posErr[:,0]*1000)
        # plt.xlabel("fit x err (micron)")

        # plt.figure()
        # plt.hist(posErr[:,1]*1000)
        # plt.xlabel("fit y err (micron)")

        plt.figure()
        plt.title("%s Fit\nRMS error %.2f microns"%(siteName, rmsErr))
        plt.hist(numpy.sqrt(posErr[:,0]**2+posErr[:,1]**2)*1000)
        plt.xlabel("fit err (micron)")

        plt.figure(figsize=(8,8))
        plt.title(
            "%s Fit Residuals\nRMS error %.2f microns"%(siteName, rmsErr)
        )
        plt.quiver(xWok,yWok,posErr[:,0], posErr[:,1], angles="xy")
        plt.xlabel("x wok (mm)")
        plt.ylabel("y wok (mm)")
        plt.axis("equal")
        plt.savefig("%sfitErrs.png"%siteName, dpi=150)

    # run the reverse
    ra, dec, fieldWarn = wokxy2radec(
        xWok, yWok, plateData["fiberType"], plateData["raCen"],
        plateData["decCen"], 0, siteName, timeObs
    )

    sk1 = SkyCoord(ra=ra * u.deg, dec=dec * u.deg)
    sk2 = SkyCoord(ra=plateData["ra"] * u.deg, dec=plateData["dec"] * u.deg)

    angSep = sk1.separation(sk2)
    asec = numpy.array(angSep)*3600
    assert numpy.max(asec) < 0.5  # less than 0.5 arcsecs round trip
    if plot:
        plt.figure()
        plt.hist(asec)
        plt.title("angular sep (arcsec)")


def test_utils():
    run_field("APO", plot=False)

    # TODO: run this when we figure out how to switch calibration files.
    # run_field("LCO")

if __name__ == "__main__":
    test_scales()
    # plt.show()



from __future__ import annotations

import pathlib
import re
import warnings
from dataclasses import dataclass
from typing import Any

import numpy
import numpy.typing as nt
import pandas
import scipy.ndimage
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS, FITSFixedWarning
from astropy.wcs.utils import fit_wcs_from_points
from matplotlib import pyplot as plt
from scipy.spatial import KDTree
from scipy.stats import sigmaclip
from skimage.registration import phase_cross_correlation
from skimage.transform import SimilarityTransform, EuclideanTransform

from coordio.conv import guideToTangent, tangentToWok

from . import Site, calibration, defaults
from .astrometry import astrometrynet_quick
from .conv import tangentToGuide, wokToTangent
from .coordinate import Coordinate, Coordinate2D
from .exceptions import CoordIOError, CoordIOUserWarning
from .extraction import sextractor_quick
from .sky import ICRS, Observed
from .telescope import Field, FocalPlane
from .utils import radec2wokxy, wokxy2radec, gaia_mags2sdss_gri
from .wok import Wok
from .transforms import arg_nearest_neighbor


__all__ = [
    "Guide",
    "GuiderFitter",
    "umeyama",
    "radec_to_gfa",
    "gfa_to_radec",
    "cross_match",
]


warnings.filterwarnings("ignore", module="astropy.wcs.wcs")
warnings.filterwarnings("ignore", category=FITSFixedWarning)


class Guide(Coordinate2D):
    """Guide coordinates are 2D cartesian xy coordinates.  Units are pixels.
    The origin is the lower left corner of the lower left pixel when looking
    at the CCD through the camera window.  Thus guide images are taken in
    "selfie-mode".  So the coordinate (0.5, 0.5) is the center of the lower
    left pixel. -y points (generally) toward the wok vertex.  True rotations
    and locations of GFAs in the wok are measured/calibrated after
    installation and on-sky characterization.

    Parameters
    -----------
    value : numpy.ndarray
        A Nx2 array where [x,y] are columns.  Or `.Tangent`
    xBin : int
        CCD x bin factor, defaults to 1
    yBin : int
        CCD y bin factor, defaults to 1

    Attributes
    -----------
    guide_warn : numpy.ndarray
        boolean array indicating coordinates outside the CCD FOV
    """

    __extra_params__ = ["xBin", "yBin"]
    __warn_arrays__ = ["guide_warn"]

    xBin: int
    yBin: int
    guide_warn: numpy.ndarray

    def __new__(cls, value, **kwargs):
        xBin = kwargs.get("xBin", None)
        if xBin is None:
            kwargs["xBin"] = 1

        yBin = kwargs.get("yBin", None)
        if yBin is None:
            kwargs["yBin"] = 1

        if isinstance(value, Coordinate):
            if value.coordSysName == "Tangent":
                # going from 3D -> 2D coord sys
                # reduce dimensionality
                arrInit = numpy.zeros((len(value), 2))
                obj = super().__new__(cls, arrInit, **kwargs)
                obj._fromTangent(value)
            else:
                raise CoordIOError(
                    "Cannot convert to Guide from %s" % value.coordSysName
                )
        else:
            obj = super().__new__(cls, value, **kwargs)
            obj._fromRaw()

        return obj

    def _fromTangent(self, tangentCoords):
        """Convert from tangent coordinates to guide coordinates"""
        if tangentCoords.holeID not in calibration.VALID_GUIDE_IDS:
            raise CoordIOError(
                "Cannot convert from wok hole %s to Guide" % tangentCoords.holeID
            )
        # make sure the guide wavelength is specified for all coords
        # this may be a little too extreme of a check
        badWl = numpy.sum(tangentCoords.wavelength - defaults.INST_TO_WAVE["GFA"]) != 0
        if badWl:
            raise CoordIOError(
                "Cannont convert to Guide coords from non-guide wavelength"
            )

        # use coords projected to tangent xy plane
        # could almost equivalently use tangentCoords[:,0:2]
        # and we may want to ditch the projection eventually if its
        # unnecessary

        # note, moved this calculation into
        # conv.tangentToGuide, may want to just use that one
        xT = tangentCoords.xProj
        yT = tangentCoords.yProj

        xPix = (1 / self.xBin) * (
            defaults.MICRONS_PER_MM / defaults.GFA_PIXEL_SIZE * xT
            + defaults.GFA_CHIP_CENTER
        )

        yPix = (1 / self.yBin) * (
            defaults.MICRONS_PER_MM / defaults.GFA_PIXEL_SIZE * yT
            + defaults.GFA_CHIP_CENTER
        )

        self[:, 0] = xPix
        self[:, 1] = yPix

        self._fromRaw()

    def _fromRaw(self):
        """Find coords that may be out of the FOV and tag them"""

        # note this might be wrong when xBin is 3
        maxX = defaults.GFA_CHIP_CENTER * 2 / self.xBin
        maxY = defaults.GFA_CHIP_CENTER * 2 / self.yBin
        xPix = self[:, 0]
        yPix = self[:, 1]

        arg = numpy.argwhere((xPix < 0) | (xPix > maxX) | (yPix < 0) | (yPix > maxY))

        if len(arg) == 0:
            # everything in range...
            return

        arg = arg.squeeze()
        self.guide_warn[arg] = True


def gfa_to_wok(observatory: str, x_pix: float, y_pix: float, gfa_id: int):
    """Converts from a GFA pixel to wok coordinates."""

    gfa_row = calibration.gfaCoords.loc[(observatory, gfa_id), :]

    b = gfa_row[["xWok", "yWok", "zWok"]].to_numpy().squeeze()
    iHat = gfa_row[["ix", "iy", "iz"]].to_numpy().squeeze()
    jHat = gfa_row[["jx", "jy", "jz"]].to_numpy().squeeze()
    kHat = gfa_row[["kx", "ky", "kz"]].to_numpy().squeeze()

    xt, yt = guideToTangent(x_pix, y_pix)
    zt = 0

    return tangentToWok(xt, yt, zt, b, iHat, jHat, kHat)  # type: ignore


def gfa_to_radec(
    observatory: str,
    x_pix: float,
    y_pix: float,
    gfa_id: int,
    bore_ra: float,
    bore_dec: float,
    position_angle: float = 0,
    offset_ra: float = 0,
    offset_dec: float = 0,
    offset_pa: float = 0,
    obstime: float | None = None,
    icrs: bool = False,
):
    """Converts from a GFA pixel to observed RA/Dec. Offsets are in arcsec."""

    site = Site(observatory)
    site.set_time(obstime)

    wavelength = defaults.INST_TO_WAVE["GFA"]

    wok_coords = gfa_to_wok(observatory, x_pix, y_pix, gfa_id)

    bore_ra += offset_ra / 3600.0 / numpy.cos(numpy.radians(bore_dec))
    bore_dec += offset_dec / 3600.0
    position_angle -= offset_pa / 3600.0

    boresight_icrs = ICRS([[bore_ra, bore_dec]])
    boresight = Observed(
        boresight_icrs,
        site=site,
        wavelength=wavelength,
    )

    wok = Wok([wok_coords], site=site, obsAngle=position_angle)
    focal = FocalPlane(wok, wavelength=wavelength, site=site)
    field = Field(focal, field_center=boresight)
    observed = Observed(field, wavelength=wavelength, site=site)

    if icrs is False:
        return (observed.ra[0], observed.dec[0])
    else:
        return ICRS(observed)[0]


def radec_to_gfa(
    observatory: str,
    ra: float | numpy.ndarray,
    dec: float | numpy.ndarray,
    gfa_id: int,
    bore_ra: float,
    bore_dec: float,
    position_angle: float = 0,
    offset_ra: float = 0,
    offset_dec: float = 0,
    offset_pa: float = 0,
    obstime: float | None = None,
):
    """Converts from observed RA/Dec to GFA pixel, taking into account offsets."""

    ra = numpy.atleast_1d(ra)
    dec = numpy.atleast_1d(dec)

    site = Site(observatory)
    site.set_time(obstime)
    assert site.time

    bore_ra += offset_ra / 3600.0 / numpy.cos(numpy.radians(bore_dec))
    bore_dec += offset_dec / 3600.0
    position_angle -= offset_pa / 3600.0

    xwok, ywok, *_ = radec2wokxy(
        ra,
        dec,
        None,
        "GFA",
        bore_ra,
        bore_dec,
        position_angle,
        observatory,
        site.time.jd,
        focalScale=1.0,  # Guider scale is relative to 1
    )

    zwok = numpy.array([defaults.POSITIONER_HEIGHT] * len(xwok))

    gfa_coords = calibration.gfaCoords.loc[observatory]
    gfa_coords.reset_index(inplace=True)

    gfa_row = gfa_coords[gfa_coords.id == gfa_id]

    b = gfa_row[["xWok", "yWok", "zWok"]].to_numpy().squeeze()
    iHat = gfa_row[["ix", "iy", "iz"]].to_numpy().squeeze()
    jHat = gfa_row[["jx", "jy", "jz"]].to_numpy().squeeze()
    kHat = gfa_row[["kx", "ky", "kz"]].to_numpy().squeeze()

    xt, yt, _ = wokToTangent(xwok, ywok, zwok, b, iHat, jHat, kHat)
    x_ccd, y_ccd = tangentToGuide(xt, yt)

    return x_ccd, y_ccd


def umeyama(X, Y):
    """Rigid alignment of two sets of points in k-dimensional Euclidean space.

    Given two sets of points in correspondence, this function computes the
    scaling, rotation, and translation that define the transform TR that
    minimizes the sum of squared errors between TR(X) and its corresponding
    points in Y.  This routine takes O(n k^3)-time.

    Parameters
    ----------
    X
        A ``k x n`` matrix whose columns are points.
    Y
        A ``k x n`` matrix whose columns are points that correspond to the
        points in X

    Returns
    -------
    c,R,t
        The  scaling, rotation matrix, and translation vector defining the
        linear map TR as ::

            TR(x) = c * R * x + t

        such that the average norm of ``TR(X(:, i) - Y(:, i))`` is
        minimized.

    Copyright: Carlo Nicolini, 2013

    Code adapted from the Mark Paskin Matlab version from
    http://openslam.informatik.uni-freiburg.de/data/svn/tjtf/trunk/matlab/ralign.m

    See paper by Umeyama (1991)

    """

    m, n = X.shape

    mx = X.mean(1)
    my = Y.mean(1)

    Xc = X - numpy.tile(mx, (n, 1)).T
    Yc = Y - numpy.tile(my, (n, 1)).T

    sx = numpy.mean(numpy.sum(Xc * Xc, 0))

    Sxy = numpy.dot(Yc, Xc.T) / n

    U, D, V = numpy.linalg.svd(Sxy, full_matrices=True, compute_uv=True)
    V = V.T.copy()

    r = numpy.linalg.matrix_rank(Sxy)
    S = numpy.eye(m)

    if r < (m - 1):
        raise CoordIOError("Not enough independent measurements")

    if numpy.linalg.det(Sxy) < 0:
        S[-1, -1] = -1
    elif r == m - 1:
        if numpy.linalg.det(U) * numpy.linalg.det(V) < 0:
            S[-1, -1] = -1

    R = numpy.dot(numpy.dot(U, S), V.T)
    c = numpy.trace(numpy.dot(numpy.diag(D), S)) / sx
    t = my - c * numpy.dot(R, mx)

    return c, R, t


def cross_match(
    measured_xy: numpy.ndarray,
    reference_xy: numpy.ndarray,
    reference_radec: numpy.ndarray,
    x_size: int,
    y_size: int,
    cross_corrlation_shift: bool = True,
    blur: float = 5,
    distance_upper_bound: int = 10,
    **kwargs,
):
    """Determines the shift between two sets of points using cross-correlation.

    Constructs 2D images from reference and measured data sets and calculates
    the image shift using cross-correlation registration. It then associated
    reference to measured detections using nearest neighbours and builds a
    WCS using the reference on-sky positions.

    Parameters
    ----------
    measured_xy
        A 2D array with the x and y coordinates of each measured point.
    reference_xy
        A 2D array with the x and y coordinates of each reference point.
    reference_radec
        A 2D array with the ra and dec coordinates of each reference point.
    x_size
        Size of the image for 2D cross-correlation.
    y_size
        Size of the image for 2D cross-correlation.
    blur
        The sigma, in pixels, of the Gaussian kernel used to convolve the images.
    distance_upper_bound
        Maximum distance, in pixels, for KD tree nearest neighbours.
    kwargs
        Other arguments to pass to ``skimage.registration.phase_cross_correlation``.

    Returns
    -------
    wcs
        A tuple with the WCS of the solution and the translation invariant normalized
        RMS error between reference and moving image (see ``phase_cross_correlation``).

    """

    # Create and blur the reference and measured images.
    ref_image = numpy.zeros((y_size, x_size), numpy.float32)
    ref_image[reference_xy.astype(int)[:, 1], reference_xy.astype(int)[:, 0]] = 1
    if blur > 0:
        ref_image = scipy.ndimage.gaussian_filter(ref_image, blur)

    meas_image = numpy.zeros((y_size, x_size), numpy.float32)
    meas_image[measured_xy.astype(int)[:, 1], measured_xy.astype(int)[:, 0]] = 1
    if blur > 0:
        meas_image = scipy.ndimage.gaussian_filter(meas_image, blur)

    # Calculate the shift and error.
    error: float
    if cross_corrlation_shift:
        shift, error, _ = phase_cross_correlation(ref_image, meas_image, **kwargs)
    else:
        shift = numpy.array([0.0, 0.0])
        error = numpy.nan

    # Apply shift.
    measured_shift = measured_xy + shift[::-1]

    # Associate measured to reference using KD tree.
    tree = KDTree(reference_xy)
    dd, ii = tree.query(measured_shift, k=1, distance_upper_bound=distance_upper_bound)

    # Reject measured objects without a close neighbour.
    # KDTree.query() assigns an index larger than the initial set of points when
    # the closest neighbour has distance > distance_upper_bound.
    valid = dd < len(reference_xy)
    ii_valid = ii[valid]

    # Select valid measured object and their corresponding RA/Dec references.
    measured_xy_valid = measured_xy[valid]
    reference_radec_valid = reference_radec[ii_valid, :]

    if len(reference_radec_valid) < 3 or len(measured_xy_valid) < 3:
        return None, 0

    # Build the WCS.
    reference_skycoord_valid = SkyCoord(
        ra=reference_radec_valid[:, 0],
        dec=reference_radec_valid[:, 1],
        unit="deg",
    )

    wcs = fit_wcs_from_points(
        (measured_xy_valid[:, 0], measured_xy_valid[:, 1]),
        reference_skycoord_valid,
    )

    assert isinstance(wcs, WCS)

    return wcs, error


@dataclass
class GuiderFit:
    """Results of a guider data fit."""

    cameras: list[int]
    gfa_wok: pandas.DataFrame
    astro_wok: pandas.DataFrame
    coeffs: list[float]
    delta_ra: float
    delta_dec: float
    delta_rot: float
    delta_scale: float
    xrms: float
    yrms: float
    rms: float
    rms_data: pandas.DataFrame
    fit: pandas.DataFrame
    fit_rms: pandas.DataFrame
    only_radec: bool = False


class GuiderFitter:
    """Fits guider data, returning the measured offsets."""

    def __init__(self, observatory: str):
        self.observatory = observatory.upper()

        self.plate_scale = defaults.PLATE_SCALE[self.observatory]
        self.pixel_size = defaults.GFA_PIXEL_SIZE
        self.pixel_scale = 1.0 / self.plate_scale * self.pixel_size * 3600 / 1000

        self.reference_pixel: tuple[int, int] = (1024, 1024)

        self.gfa_wok_coords = self._gfa_to_wok()
        self.astro_data: pandas.DataFrame | None = None

        self.result: GuiderFit | None = None

    def reset(self):
        """Clears the astrometric data."""

        self.astro_data = None
        self.result = None

    def _gfa_to_wok(self):
        """Converts GFA reference points to wok coordinates."""

        if self.observatory not in calibration.gfaCoords.index.get_level_values(0):
            raise CoordIOError(
                f"GFA coordinates for {self.observatory} are missing. "
                "Did you load the correct calibrations?"
            )

        gfaCoords = calibration.gfaCoords.loc[self.observatory]

        wok_coords = {}

        for gfa_id in range(1, 7):
            camera_coords: pandas.DataFrame = gfaCoords.loc[gfa_id]

            b = camera_coords[["xWok", "yWok", "zWok"]].to_numpy().squeeze()
            iHat = camera_coords[["ix", "iy", "iz"]].to_numpy().squeeze()
            jHat = camera_coords[["jx", "jy", "jz"]].to_numpy().squeeze()
            kHat = camera_coords[["kx", "ky", "kz"]].to_numpy().squeeze()

            xt, yt = guideToTangent(self.reference_pixel[0], self.reference_pixel[1])
            zt = 0

            wok_coords[gfa_id] = tangentToWok(xt, yt, zt, b, iHat, jHat, kHat)

        df = pandas.DataFrame.from_dict(wok_coords).transpose()
        df.index.name = "gfa_id"
        df.columns = ["xwok", "ywok", "zwok"]

        return df

    def add_astro(
        self,
        camera: str | int,
        ra: float,
        dec: float,
        obstime: float,
        x_pixel: float,
        y_pixel: float,
    ):
        """Adds an astrometric measurement."""

        if isinstance(camera, str):
            match = re.match(r".*([1-6]).*", camera)
            if not match:
                raise CoordIOError(f"Cannot understand camera {camera!r}.")
            camera = int(match.group(1))

        new_data = (camera, ra, dec, obstime, x_pixel, y_pixel)

        if self.astro_data is None:
            self.astro_data = pandas.DataFrame(
                [new_data],
                columns=["gfa_id", "ra", "dec", "obstime", "x", "y"],
            )
        else:
            self.astro_data.loc[len(self.astro_data.index)] = new_data

    def add_wcs(
        self,
        camera: str | int,
        wcs: WCS,
        obstime: float,
        pixels: numpy.ndarray | None = None,
    ):
        """Adds a camera measurement from a WCS solution."""

        if pixels is None:
            coords: Any = wcs.pixel_to_world(*self.reference_pixel)

            ra = coords.ra.value
            dec = coords.dec.value

            self.add_astro(camera, ra, dec, obstime, *self.reference_pixel)

        else:
            ras, decs = wcs.all_pix2world(pixels[:, 0], pixels[:, 1], 0)
            for ii in range(len(ras)):
                self.add_astro(
                    camera,
                    ras[ii],
                    decs[ii],
                    obstime,
                    pixels[ii, 0],
                    pixels[ii, 1],
                )

    def add_gimg(self, path: str | pathlib.Path, pixels: numpy.ndarray | None = None):
        """Processes a raw GFA image, runs astrometry.net, adds the solution."""

        data = fits.getdata(str(path))
        header = fits.getheader(str(path), 1)

        if header["OBSERVAT"] != self.observatory:
            raise CoordIOError("GFA image is from a different observatory.")

        if "RAFIELD" not in header or "DECFIELD" not in header:
            raise CoordIOError("GFA image does not have RAFIELD or DECFIELD.")

        ra_field = header["RAFIELD"]
        dec_field = header["DECFIELD"]
        pa_field = header["FIELDPA"]

        sources = sextractor_quick(data)

        camera_id = int(header["CAMNAME"][3])
        radec_centre = gfa_to_radec(
            self.observatory,
            self.reference_pixel[0],
            self.reference_pixel[1],
            camera_id,
            ra_field,
            dec_field,
            position_angle=pa_field,
        )

        wcs = astrometrynet_quick(
            sources,
            radec_centre[0],
            radec_centre[1],
            self.pixel_scale,
        )

        if wcs is None:
            raise CoordIOError("astrometry.net could not solve image.")

        obstime = Time(header["DATE-OBS"], format="iso").jd

        self.add_wcs(camera_id, wcs, obstime, pixels=pixels)

        return wcs

    def add_proc_gimg(
        self,
        path: str | pathlib.Path,
        pixels: numpy.ndarray | None = None,
    ):
        """Adds an astrometric solution from a ``proc-gimg`` image."""

        hdu = fits.open(str(path))
        header = hdu[1].header

        is_proc = len(hdu) > 1 and "SOLVED" in header
        if not is_proc:
            raise CoordIOError(f"{path!s} is not a proc-gimg image.")

        if header["OBSERVAT"] != self.observatory:
            raise CoordIOError("GFA image is from a different observatory.")

        if not header.get("SOLVED", False):
            raise CoordIOError(f"Image {path!s} has not been astrometrically solved.")

        wcs = WCS(header)
        obstime = Time(header["DATE-OBS"], format="iso").jd

        self.add_wcs(header["CAMNAME"], wcs, obstime, pixels=pixels)

    def fit(
        self,
        field_ra: float,
        field_dec: float,
        field_pa: float = 0.0,
        offset: tuple[float, float, float] | numpy.ndarray = (0.0, 0.0, 0.0),
        scale_rms: bool = False,
        only_radec: bool = False,
        cameras: list[int] | None = None,
    ):
        """Performs the fit and returns a `.GuiderFit` object.

        The fit is performed using `.umeyama`.

        Parameters
        ----------
        field_ra
            The field centre RA.
        field_dec
            The field centre declination.
        field_pa
            The field centre position angle.
        offset
            The offset in RA, Dec, PA to apply.
        scale_rms
            Whether to correct the RMS using the measured scale factor.
        only_radec
            If `True`, fits only translation. Useful when only one or two cameras
            are solving. In this case the rotation and scale offsets will still be
            calculated using the Umeyama model, but the ra and dec offsets are
            overridden with a simple translation. The `.GuiderFit` object will
            have `only_radec=True`.

        """

        offset_ra, offset_dec, offset_pa = offset

        if self.astro_data is None:
            raise CoordIOError("Astro data has not been set.")

        camera_ids: list[int] = []
        xwok_gfa: list[float] = []
        ywok_gfa: list[float] = []
        xwok_astro: list[float] = []
        ywok_astro: list[float] = []
        use_detections: list[bool] = []
        used_cameras: set[int] = set([])

        for d in self.astro_data.itertuples():
            camera_id: int = int(d.gfa_id)
            if cameras is None or camera_id in cameras:
                used_cameras.add(camera_id)

            camera_ids.append(camera_id)

            x_wok, y_wok, _ = gfa_to_wok(self.observatory, d.x, d.y, camera_id)
            xwok_gfa.append(x_wok)
            ywok_gfa.append(y_wok)

            _xwok_astro, _ywok_astro, *_ = radec2wokxy(
                [d.ra],
                [d.dec],
                None,
                "GFA",
                field_ra - offset_ra / numpy.cos(numpy.deg2rad(d.dec)) / 3600.0,
                field_dec - offset_dec / 3600.0,
                field_pa - offset_pa / 3600.0,
                self.observatory,
                d.obstime,
                focalScale=1.0,  # Guider scale is relative to 1
            )

            xwok_astro += _xwok_astro.tolist()
            ywok_astro += _ywok_astro.tolist()

            # Use these detections fit unless we are excluding the camera.
            use_detections.append(cameras is None or camera_id in cameras)

        # Create arrays with all detections, then with only selected ones.
        X_all: nt.NDArray[numpy.float32] = numpy.array([xwok_gfa, ywok_gfa])
        Y_all: nt.NDArray[numpy.float32] = numpy.array([xwok_astro, ywok_astro])
        X_fit = X_all[:, use_detections]  # GFA to wok coordinates as a 2D array
        Y_fit = Y_all[:, use_detections]  # ICRS to wok coordinates as a 2D array

        try:
            c, R, t = umeyama(X_fit, Y_fit)
        except ValueError:
            if only_radec is True:
                warnings.warn(
                    "Cannot fit using Umeyama. Assuming unitary rotation and scale.",
                    CoordIOUserWarning,
                )
                c = 1.0
                R = numpy.array([[1, 0], [1, 0]])
                t = numpy.array([0.0, 0.0])
            else:
                return False

        if only_radec is True:
            # Override the translation component with a simple average translation
            t = (Y_fit - X_fit).mean(axis=1).T

        coeffs = [c, R, t]

        # delta_x and delta_y only align with RA/Dec if PA=0. Otherwise we need to
        # project using the PA.
        pa_rad = numpy.deg2rad(field_pa - offset_pa / 3600.0)
        delta_ra = t[0] * numpy.cos(pa_rad) + t[1] * numpy.sin(pa_rad)
        delta_dec = -t[0] * numpy.sin(pa_rad) + t[1] * numpy.cos(pa_rad)

        # Convert to arcsec and round up
        delta_ra = float(numpy.round(delta_ra / self.plate_scale * 3600.0, 3))
        delta_dec = float(numpy.round(delta_dec / self.plate_scale * 3600.0, 3))

        delta_rot = -numpy.rad2deg(numpy.arctan2(R[1, 0], R[0, 0])) * 3600.0
        delta_rot = float(numpy.round(delta_rot, 1))

        delta_scale = float(numpy.round(c, 6))

        detection_id = numpy.arange(len(xwok_astro)) + 1

        astro_wok = pandas.DataFrame.from_records(
            {
                "gfa_id": camera_ids,
                "detection_id": detection_id,
                "xwok": xwok_astro.copy(),
                "ywok": ywok_astro.copy(),
                "selected": numpy.array(use_detections).astype(int),
            },
            index=["gfa_id", "detection_id"],
        )

        gfa_wok = pandas.DataFrame.from_records(
            {
                "gfa_id": camera_ids,
                "detection_id": detection_id,
                "xwok": xwok_gfa.copy(),
                "ywok": ywok_gfa.copy(),
                "selected": numpy.array(use_detections).astype(int),
            },
            index=["gfa_id", "detection_id"],
        )

        rms_df = self.calculate_rms(
            gfa_wok,
            astro_wok,
            scale=c if scale_rms else 1,
            cameras=cameras,
        )

        fit_data = c * R @ X_all + t[numpy.newaxis].T
        fit_df = pandas.DataFrame.from_records(
            {
                "gfa_id": camera_ids,
                "detection_id": detection_id,
                "xwok": fit_data[0, :],
                "ywok": fit_data[1, :],
                "selected": numpy.array(use_detections).astype(int),
            },
            index=["gfa_id", "detection_id"],
        )
        fit_rms_df = self.calculate_rms(fit_df, astro_wok, cameras=cameras)

        self.result = GuiderFit(
            list(used_cameras),
            gfa_wok,
            astro_wok,
            coeffs,
            delta_ra,
            delta_dec,
            delta_rot,
            delta_scale,
            rms_df.loc[0].xrms,
            rms_df.loc[0].yrms,
            rms_df.loc[0].rms,
            rms_df,
            fit_df,
            fit_rms_df,
            only_radec=only_radec,
        )

        return self.result

    def calculate_rms(
        self,
        gfa_wok: pandas.DataFrame,
        astro_wok: pandas.DataFrame,
        scale: float = 1,
        cameras: list[int] | None = None,
    ):
        """Calculates the RMS of the measurements.

        Parameters
        ----------
        gfa_wok
            GFA to wok coordinates data. The data frame contains three columns:
            ``gfa_id``, ``xwok``, and ``ywok``.
        astro_wok
            Astrometric solution to wok coordinates data. The data frame contains
            three columns: ``gfa_id``, ``xwok``, and ``ywok``.
        scale
            Factor by which to scale coordinates.
        cameras
            The cameras to use to calculate the global RMS.

        Returns
        -------
        rms
            A data frame with columns ``gfa_id``, ``xrms``, ``yrms``, ``rms``
            for the x, y, and combined RMS measurements for each camera. An
            additional ``gfa_id=0`` is added with the RMS measurements for all
            GFA cameras combined. RMS values are returned in mm on wok coordinates.

        """

        def calc_rms(gfa_cam: pandas.DataFrame):
            gfa_id = gfa_cam.index.values[0][0]
            astro_cam = astro_wok.loc[gfa_id]

            delta = gfa_cam - astro_cam
            xrms = numpy.sqrt(numpy.mean(delta.xwok**2))
            yrms = numpy.sqrt(numpy.mean(delta.ywok**2))
            rms = numpy.sqrt(numpy.mean(delta.xwok**2 + delta.ywok**2))

            return pandas.Series([xrms, yrms, rms], index=["xrms", "yrms", "rms"])

        astro_wok = astro_wok.copy()
        gfa_wok = gfa_wok.copy()
        gfa_wok.loc[:, ["xwok", "ywok"]] *= scale

        rms_df = gfa_wok.groupby("gfa_id").apply(calc_rms)

        if cameras is None:
            delta = gfa_wok - astro_wok
        else:
            delta = gfa_wok.loc[cameras] - astro_wok.loc[cameras]
        xrms = numpy.sqrt(numpy.mean(delta.xwok**2))
        yrms = numpy.sqrt(numpy.mean(delta.ywok**2))
        rms = numpy.sqrt(numpy.mean(delta.xwok**2 + delta.ywok**2))

        rms_df.loc[0, :] = (xrms, yrms, rms)

        return rms_df

    def plot_fit(self, filename: str | pathlib.Path):
        """Plot the fit results."""

        if self.result is None:
            raise RuntimeError("fit() needs to be run before plot_fit().")

        fig, ax = plt.subplots(2, 3)
        fig.tight_layout()

        iax = 0
        jax = 0
        direction = 1
        for cam_id in range(1, 7):
            one_arcsec = self.plate_scale / 3600.0

            camera_note = None
            color = "k"

            if cam_id in self.result.fit.index:
                X = self.result.fit.loc[cam_id].xwok
                Y = self.result.fit.loc[cam_id].ywok

                U = self.result.astro_wok.loc[cam_id].xwok
                V = self.result.astro_wok.loc[cam_id].ywok
                U = U - X
                V = V - Y

                rms_mm = float(self.result.fit_rms.loc[cam_id].rms)
                rms = numpy.round(rms_mm / self.plate_scale * 3600, 4)

                if self.result.cameras and cam_id not in self.result.cameras:
                    camera_note = "Not fit"
                    color = "r"

            else:
                X = Y = U = V = []
                rms = -999
                camera_note = "Disabled / not solved"
                color = "r"

            q = ax[iax][jax].quiver(
                X,
                Y,
                U,
                V,
                color=color,
                units="xy",
                angles="xy",
                scale_units="inches",
                scale=0.15,
            )

            if cam_id == 1:
                ax[iax][jax].quiverkey(q, 0.2, 0.9, one_arcsec, label="1 arcsec")

            text1 = ax[iax][jax].text(
                0.5,
                0.95,
                f"RMS: {rms}",
                color=color,
                horizontalalignment="center",
                verticalalignment="center",
                transform=ax[iax][jax].transAxes,
            )
            text1.set_bbox(dict(facecolor="white", alpha=1, edgecolor="none"))

            if camera_note:
                text2 = ax[iax][jax].text(
                    0.5,
                    0.90,
                    camera_note,
                    color=color,
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=ax[iax][jax].transAxes,
                )
                text2.set_bbox(dict(facecolor="white", alpha=1, edgecolor="none"))

            ax[iax][jax].set_title(f"GFA{cam_id}", color=color)

            jax += direction
            if jax == 3:
                jax = -1
                iax = 1
                direction = -1

        fig.savefig(str(filename))


class SolvePointing:
    """
    Determines the best-fit telescope boresight pointing from a set of GFA
    images.  Requires specifying a guess or reference pointing as a starting
    point.  The reference pointing can be assumed from the gfa image headers.

    Parameters
    -------------
    raCen : float | None = None
        RA of field center in degrees.  If None, user must specify pt_source
    decCen : float | None = None
        Dec of field center in degrees.  If None, user must specify pt_source
    paCen : float | None = None
        PA of field center in degrees, If None, user must specify pt_src
    pt_source : str | None = None
        either "design" or "telescope".  Design will solve using robostrategy
        intented field center, telescope will solve using the reported
        sky position of the telescope boresight. Data pulled from gcam headers.
        If raCen, decCen, and/or paCen are specified, those values are used
        instead of data from image headers
    offset_ra : float = 0
        offset in ra to add to the supplied field center (true arcseconds)
    offset_dec : float = 0
        offset in dec to add to the supplied field center (arcseconds)
    offset_pa : float = 0
        offset in pa to subtract from supplied field angle (arcseconds)
    scale : float | None = None
        scale factor for the field
    db_conn_st : str
        connection string for the db
    db_tab_name : str
        table in database with gaia info
    """
    def __init__(
        self,
        raCen: float | None = None,
        decCen: float | None = None,
        paCen: float | None = None,
        scale: float | None = None,
        pt_source: str | None = None,
        offset_ra: float = 0,
        offset_dec: float = 0,
        offset_pa: float = 0,
        db_conn_st: str = "postgresql://sdss_user@operations.sdss.org/sdss5db",
        db_tab_name: str = "catalogdb.gaia_dr2_source_g_lt_18"
    ):

        if pt_source is None and None in [raCen, decCen, paCen]:
            raise RuntimeError(
                "must specify either pt_source or raCen, decCen, paCen"
            )
        if raCen is None or decCen is None or paCen is None:
            if pt_source not in ["telescope", "design"]:
                raise RuntimeError(
                    "pt_source must be either 'telescope' or 'design'"
                )

        self._raCen = raCen
        self._decCen = decCen
        self._paCen = paCen
        self.pt_source = pt_source
        self.offset_ra = offset_ra
        self.offset_dec = offset_dec
        self.offset_pa = offset_pa
        self.scale = scale
        self.db_conn_st = db_conn_st
        self.db_tab_name = db_tab_name

        self.GAIA_EPOCH = 2457206

        # populated by add_gimg
        self.fieldInit = False
        self.observatory = None
        self.raCenRef = None
        self.decCenRef = None
        self.paCenRef = None
        self.obsTimeRef = None
        self.ipa = None
        self.imgNum = None
        self.gfaHeaders = {}
        self.gfaWCS = {}
        self.allCentroids = pandas.DataFrame()

        # populated or modified by solve
        self.allGaia = pandas.DataFrame()
        self.skyDistThresh = None  # arcseconds for a match hit
        self.raCenMeas = None
        self.decCenMeas = None
        self.paCenMeas = None
        self.scaleMeas = None
        self.altCenMeas = None
        self.azCenMeas = None
        # self.fit_rms = None
        self.matchedSources = pandas.DataFrame()
        self.translation = numpy.array([0.0, 0.0])
        self.rotation = 0.0

    def getMetadata(self):
        """Get a list of data that can be easily stuffed in a fits
        header.
        """

        metaDataList = [
            ("SOL_RA", self.raCenMeas, "solved RA of pointing (deg)"),
            ("SOL_DEC", self.decCenMeas, "solved Dec of pointing (deg)"),
            ("SOL_PA", self.paCenMeas, "solved PA of pointing (deg)"),
            ("SOL_SCL", self.scaleMeas, "solved scale factor"),
            ("SOL_ALT", self.altCenMeas, "solved alt of pointing (deg)"),
            ("SOL_AZ", self.azCenMeas, "solved az of pointing (deg)"),
            ("SOL_GRMS", self.guide_rms, "guide rms between reference and solved (mm)"),
            ("SOL_FRMS", self.fit_rms, "rms of fit (mm)"),
            ("REF_RA", self.raCenRef, "reference RA (deg)"),
            ("REF_DEC", self.decCenRef, "reference Dec (deg)"),
            ("REF_PA", self.paCenRef, "reference PA (deg)"),
            ("REF_SCL", self.scale, "reference scale factor"),
            ("TAI_MID", self.obsTimeRef.mjd * 24 * 60 * 60, "tai seconds at middle of exposure"),
            # ("N_WCS", len(self.gfaWCS), "number of GFAs with astronet solutions"),
            ("N_STARS", len(self.matchedSources), "number of stars used in fit"),
            ("N_GFAS", len(set(self.matchedSources.gfaNum)), "number of GFAs used in fit"),
            ("NITR_WCS", self.nIterWCS, "number of iterations for wcs solve"),
            ("NITR_ALL", self.nIterAll, "number of iterations for full solve")
        ]

        return metaDataList

    def median_zeropoint(self, gfaNum):
        _df = self.matchedSources[self.matchedSources.gfaNum == gfaNum]
        return numpy.nanmedian(_df.zp)

    @property
    def plate_scale(self):
        # mm per deg
        return defaults.PLATE_SCALE[self.observatory]

    # @property
    # def pixel_scale(self):
    #     # arcsec per pixel
    #     return 1.0 / self.plate_scale * defaults.GFA_PIXEL_SIZE * 3600 / 1000

    @property
    def wokDistThresh(self):
        # mm
        return self.skyDistThresh / 3600. * self.plate_scale

    @property
    def xyWokPredictRef(self):
        output = radec2wokxy(
            self.matchedSources.ra.to_numpy(),
            self.matchedSources.dec.to_numpy(), self.GAIA_EPOCH,
            "GFA", self.raCenRef, self.decCenRef, self.paCenRef,
            self.observatory, self.obsTimeRef.jd, focalScale=self.scaleMeas,
            pmra=self.matchedSources.pmra.to_numpy(),
            pmdec=self.matchedSources.pmdec.to_numpy(),
            parallax=self.matchedSources.parallax.to_numpy(), fullOutput=True
        )

        xPred = output[0]
        yPred = output[1]
        return xPred, yPred

    def fitWCS(self, gfaNum):
        # return a dictionary keyed by GFA ID

        _df = self.matchedSources[self.matchedSources.gfaNum==gfaNum]

        if len(_df) < 3:
            # require at least 3 stars for a wcs solution?
            return None

        xy = _df[["x", "y"]].to_numpy()
        raDec = _df[["ra", "dec"]].to_numpy()

        # Build the WCS.
        reference_skycoord_valid = SkyCoord(
            ra=raDec[:, 0],
            dec=raDec[:, 1],
            unit="deg",
        )

        wcs = fit_wcs_from_points(
            (xy[:, 0], xy[:, 1]),
            reference_skycoord_valid,
        )

        return wcs


    def rms_df(self, fit=False):
        """
        returns rms for expected vs solved pointing
        ----------
        A data frame with columns ``gfa_id``, ``xrms``, ``yrms``, ``rms``
        for the x, y, and combined RMS measurements for each camera. An
        additional ``gfa_id=0`` is added with the RMS measurements for all
        GFA cameras combined. RMS values are returned in mm on wok coordinates.
        """
        df = self.matchedSources.copy()
        if not fit:
            # reference rms
            xPred, yPred = self.xyWokPredictRef
            df["xWokPred"] = xPred
            df["yWokPred"] = yPred

        _gfa_id = []
        _xrms = []
        _yrms = []
        _rms = []

        for gfaNum in list(set(df.gfaNum)):
            _df = df[df.gfaNum==gfaNum]
            dx = _df.xWokPred - _df.xWokMeas
            dy = _df.yWokPred - _df.yWokMeas

            _gfa_id.append(gfaNum)
            _xrms.append(numpy.sqrt(numpy.mean(dx**2)))
            _yrms.append(numpy.sqrt(numpy.mean(dy**2)))
            _rms.append(numpy.sqrt(numpy.mean(dx**2+dy**2)))

        # finally add total rms under gfa_id = 0
        dx = df.xWokPred - df.xWokMeas
        dy = df.yWokPred - df.yWokMeas

        _gfa_id.append(0)
        _xrms.append(numpy.sqrt(numpy.mean(dx**2)))
        _yrms.append(numpy.sqrt(numpy.mean(dy**2)))
        _rms.append(numpy.sqrt(numpy.mean(dx**2+dy**2)))

        out = pandas.DataFrame({
            "gfa_id": _gfa_id,
            "xrms": _xrms,
            "yrms": _yrms,
            "rms": _rms,
        })
        # out.set_index("gfa_id")

        return out.set_index("gfa_id")

    @property
    def guide_rms(self):
        # RMS error between stars in reference frame and solved frame
        # the measured scale is applied to remove a scale dependence
        # on rms
        xPred, yPred = self.xyWokPredictRef

        xFit = self.matchedSources.xWokPred.to_numpy()
        yFit = self.matchedSources.yWokPred.to_numpy()

        rms = numpy.sqrt(
            numpy.mean((xPred - xFit)**2 + (yPred - yFit)**2)
        )
        return rms

    @property
    def fit_rms(self):
        return numpy.sqrt(numpy.mean(self.matchedSources.dr**2))

    @property
    def guide_rms_sky(self):
        # arcsec
        return self.guide_rms / self.plate_scale * 3600

    @property
    def fit_rms_sky(self):
        # arcsec
        return self.fit_rms / self.plate_scale * 3600

    @property
    def used_cameras(self):
        gfaNums = set(self.matchedSources.gfaNum)
        return sorted(list(gfaNums))

    @property
    def n_stars_used(self):
        return len(self.matchedSources)

    @property
    def gfa_wok(self):
        gfa_id = self.matchedSources.gfaNum.to_numpy(),
        detection_id = numpy.arange(len(gfa_id))
        x_wok = self.matchedSources.xWokMeas.to_numpy(),
        y_wok = self.matchedSources.yWokMeas.to_numpy(),
        selected = numpy.ones(len(gfa_id))
        df = pandas.DataFrame(
            {
                "gfa_id": gfa_id,
                "detection_id": detection_id,
                "xwok": x_wok,
                "ywok": y_wok,
                "selected": selected,
            },
            index=["gfa_id", "detection_id"],
        )
        return df

    @property
    def astro_wok(self):
        gfa_id = self.matchedSources.gfaNum.to_numpy()
        detection_id = numpy.arange(len(gfa_id))
        x_wok, y_wok = self.xyWokPredictRef
        selected = numpy.ones(len(gfa_id))
        df = pandas.DataFrame.from_records(
            {
                "gfa_id": gfa_id,
                "detection_id": detection_id,
                "xwok": x_wok,
                "ywok": y_wok,
                "selected": selected,
            },
            index=["gfa_id", "detection_id"],
        )
        return df

    @property
    def fit_data(self):
        gfa_id = self.matchedSources.gfaNum.to_numpy()
        detection_id = numpy.arange(len(gfa_id))
        x_wok = self.matchedSources.xWokPred.to_numpy()
        y_wok = self.matchedSources.yWokPred.to_numpy()
        selected = numpy.ones(len(gfa_id))
        df = pandas.DataFrame.from_records(
            {
                "gfa_id": gfa_id,
                "detection_id": detection_id,
                "xwok": x_wok,
                "ywok": y_wok,
                "selected": selected,
            },
            index=["gfa_id", "detection_id"],
        )
        return df

    @property
    def guider_fit(self):
        rms_df = self.rms_df()
        fit_rms_df = self.rms_df(fit=True)

        result = GuiderFit(
            list(self.used_cameras),
            self.gfa_wok,
            self.astro_wok,
            self.coeffs,
            self.delta_ra,
            self.delta_dec,
            self.delta_rot,
            self.delta_scale,
            rms_df.loc[0].xrms,
            rms_df.loc[0].yrms,
            rms_df.loc[0].rms,
            rms_df,
            self.fit_data,
            fit_rms_df,
            only_radec=False,
        )

        return result

    @property
    def coeffs(self):
        rotMat = numpy.array([
            [numpy.cos(self.rotation), -numpy.sin(self.rotation)],
            [numpy.sin(self.rotation), numpy.cos(self.rotation)]
        ])
        return [self.scaleMeas, rotMat, self.translation]

    @property
    def delta_ra(self):
        cosDec = numpy.cos(numpy.radians(self.decCenMeas))
        return (self.raCenMeas - self.raCenRef) * 3600 * cosDec

    @property
    def delta_dec(self):
        return (self.decCenMeas - self.decCenRef) * 3600

    @property
    def delta_rot(self):
        return (self.paCenMeas - self.paCenRef) * 3600

    @property
    def delta_scale(self):
        return self.scaleMeas

    def pix2wok(self, x, y, gfaNum):
        g = calibration.gfaCoords.loc[(self.observatory, gfaNum), :]
        zt = numpy.zeros(len(x))
        b = g[["xWok", "yWok", "zWok"]].to_numpy()
        iHat = g[["ix", "iy", "iz"]].to_numpy()
        jHat = g[["jx", "jy", "jz"]].to_numpy()
        kHat = g[["kx", "ky", "kz"]].to_numpy()

        xt, yt = guideToTangent(x, y)

        xw, yw, zw = tangentToWok(
            xt, yt, zt,
            b, iHat, jHat, kHat
        )
        return xw, yw

    def wok2pix(self, x, y, gfaNum):
        g = calibration.gfaCoords.loc[(self.observatory, gfaNum), :]
        zw = numpy.zeros(len(x))
        b = g[["xWok", "yWok", "zWok"]].to_numpy()
        iHat = g[["ix", "iy", "iz"]].to_numpy()
        jHat = g[["jx", "jy", "jz"]].to_numpy()
        kHat = g[["kx", "ky", "kz"]].to_numpy()

        xt, yt, zt = wokToTangent(
            x, y, zw,
            b, iHat, jHat, kHat
        )

        xyPix = tangentToGuide(xt,yt) #,y_0=float(g.y_0), y_1=float(g.y_1))
        return xyPix

    def gfa2radec(self, gfaNum):
        """ get the ra/dec of ccd center, if wcs is available use that """
        xw, yw = self.pix2wok(numpy.array([1024]), numpy.array([1024]), gfaNum)
        ra, dec, fieldWarn = wokxy2radec(
            numpy.array(xw), numpy.array(yw), "GFA", self.raCenMeas,
            self.decCenMeas, self.paCenMeas, self.observatory,
            self.obsTimeRef.jd, focalScale=self.scaleMeas
        )
        return ra[0], dec[0]

    def initializeField(self, observatory, imgNum, hdr):
        self.observatory = observatory

        if self.pt_source == "design":
            _raCen = hdr["RAFIELD"]
            _decCen = hdr["DECFIELD"]
            _paCen = hdr["FIELDPA"]
        elif self.pt_source == "telescope":
            # warning assumes APO for now
            # will need to modify for LCO
            _raCen = hdr["RA"]  # RA is ObjNetPos, RADEG is ObjPos (tcc kws)
            _decCen = hdr["DEC"]
            _paCen = hdr["ROTPOS"] # wont work for LCO no ROTPOS header

        # if no field center was specified, use field center from
        # header
        if self._raCen is None:
            self._raCen = _raCen
        if self._decCen is None:
            self._decCen = _decCen
        if self._paCen is None:
            self._paCen = _paCen

        # apply offsets if specified
        ddec = self.offset_dec / 3600.
        self.decCenRef = self._decCen - ddec
        dra = self.offset_ra / 3600. / numpy.cos(numpy.radians(self.decCenRef))
        self.raCenRef = self._raCen - dra
        self.paCenRef = self._paCen - self.offset_pa / 3600.

        # initalize the starting point for the iteration at the reference
        # pointing
        self.raCenMeas = self.raCenRef
        self.decCenMeas = self.decCenRef
        self.paCenMeas = self.paCenRef
        if self.scale is None:
            self.scale = defaults.SITE_TO_SCALE[self.observatory]
        self.scaleMeas = self.scale

        self.tStart = Time(hdr["DATE-OBS"], format="iso", scale="tai")
        self.exptime = hdr["EXPTIMEN"]
        dt = TimeDelta(self.exptime/2.0, format="sec", scale="tai")
        # choose exposure end as reference time because
        # it better matches the telescope headers (eg for a pointing model)
        self.obsTimeRef = self.tStart + dt
        # dt = TimeDelta(hdr["EXPTIME"]/2., format="sec", scale="tai")
        # self.obsTimeMid = self.tStart + dt
        # # tcc need this in seconds
        # self.taiMid = self.obsTimeRef.mjd * 24 * 60 * 60
        self.imgNum = imgNum
        self.ipa = hdr["IPA"]
        if self.observatory == "APO":
            self.fieldCenAltRef = hdr["ALT"]
            self.fieldCenAzRef = hdr["AZ"]
        else:
            self.fieldCenAltRef = None
            self.fieldCenAzRef = None
        self.initField = True

    def add_gimg(
        self,
        img_path: str | pathlib.Path,
        centroids: pandas.DataFrame | None = None,
        wcs: WCS | None = None,
        gain: float | None = None
    ):
        """
        Add gfa data to be used in fitting.  Don't add a gfa you don't want
        incuded in fitting.  Strict checking is done to make sure all images
        added have the same exposure number.  Expects SDSS style naming and
        headers, should work for both gimg*.fits and proc-gimg*.fits files.

        Parameters
        --------------
        img_path
            path to a gimg or proc-gimg file
        centroids
            sep style extracted parameters in pandas.DataFrame form.
            Additionally if a "CENTROIDS" extension is present in the fits
            file, those will be used.  If not present, centroids are extracted
            from the data.
        wcs
            a WCS object
        gain
            electrons per adu for this camera
        """
        img_path = str(img_path)
        ff = fits.open(img_path)
        if centroids is None or len(centroids)==0:
            # extract centroids
            centroids = sextractor_quick(ff[1].data)
        if len(centroids) == 0:
            # no centroids found skip this gfa
            ff.close()
            return

        hdr = dict(ff[1].header)

        tokens = img_path.strip(".fits").split("/")[-1].split("-")
        imgNum = int(tokens[-1])
        gfaNum = int(tokens[-2].strip("gfa").strip("n").strip("s"))
        self.gfaHeaders[gfaNum] = hdr

        if not self.fieldInit:
            if tokens[-2].endswith("n"):
                observatory = "APO"
            else:
                observatory = "LCO"
            self.initializeField(observatory, imgNum, hdr)

        if imgNum != self.imgNum:
            raise RuntimeError(
                "May not add gfa files with differing img numbers"
            )

        # fwhm = 2 * (numpy.log(2) * (centroids.a**2 + centroids.b**2))**0.5
        # centroids["fwhm"] = fwhm

        centroids["gfaNum"] = gfaNum
        # remove saturated sources
        # centroids = centroids[centroids.peak < 55000].reset_index(drop=True)
        # calculate wok coordinates for each centroid
        xWokMeas, yWokMeas = self.pix2wok(
            centroids.x.to_numpy(),
            centroids.y.to_numpy(),
            gfaNum
        )
        centroids["xWokMeas"] = xWokMeas
        centroids["yWokMeas"] = yWokMeas

        if wcs is not None and "CTYPE1" in wcs.to_header():
            self.gfaWCS[gfaNum] = wcs
            # use wcs to calculate on-sky locations
            # of centroids note these are off by 0.5
            # but leaving so that cherno default rotation and fudge factor
            # don't need to change
            xyCents = centroids[["x", "y"]].to_numpy()
            raDecMeas = numpy.array(wcs.pixel_to_world_values(xyCents))
            centroids["raMeas"] = raDecMeas[:, 0]
            centroids["decMeas"] = raDecMeas[:, 1]
            # import pdb; pdb.set_trace()

        centroids["gain"] = gain

        self.allCentroids = pandas.concat(
            [self.allCentroids, centroids], ignore_index=True
        )

        # if wcs is not None:
        #     self.gfaWCS[gfaNum] = WCS(open(wcs).read())

        ff.close()

    def getGaiaSources(self, ra, dec, radius=0.16, magLimit=18):
        # import time; tstart=time.time()
        query = "SELECT source_id, ra, dec, pmra, pmdec, parallax, phot_g_mean_mag, bp_rp"
        query += " FROM %s" % self.db_tab_name
        query += " WHERE q3c_radial_query"
        query += "(ra, dec, %.4f, %.4f, %.4f)" % (ra, dec, radius)
        query += " AND phot_g_mean_mag < %.2f" % magLimit
        # print(query)
        df = pandas.read_sql(query, self.db_conn_st)
        df = df.sort_values("phot_g_mean_mag")
        df = df.head(250).reset_index(drop=True)
        # print("max mag", numpy.max(df.phot_g_mean_mag))
        # print("getGaiaSources took", time.time()-tstart)
        return df.dropna().reset_index(drop=True)

        # self.allGaia = pandas.concat(allGaia)
        # print("got", len(self.allGaia), " sources")

    def _updateZP(self):
        xWokPredRef, yWokPredRef = self.xyWokPredictRef
        self.matchedSources["xWokPredRef"] = xWokPredRef
        self.matchedSources["yWokPredRef"] = yWokPredRef
        sdss_g, sdss_r, sdss_i = gaia_mags2sdss_gri(
            gaia_g=self.matchedSources.phot_g_mean_mag.to_numpy(),
            gaia_bp_rp=self.matchedSources.bp_rp.to_numpy()
        )
        self.matchedSources["sdss_r"] = sdss_r

        # when re-analyzing old images, aperflux is not available in the
        # centroids table unless you re-extract sources.  For this case, just
        # replace it with flux which has always been there.
        if "aperflux" in list(self.matchedSources.columns):
            flux = self.matchedSources.aperflux
        else:
            flux = self.matchedSources.flux

        zp = 2.5 * numpy.log10(
            flux * self.matchedSources.gain / self.exptime
        ) + sdss_r
        self.matchedSources["zp"] = zp

    def _matchWCS(self):
        """Note, only works if there was at least 1 wcs solution provided!
        """
        # allGaia = []
        # for gfaNum, wcs in self.gfaWCS.items():
        #     ra, dec = wcs.pixel_to_world_values([[1024,1024]])[0]
        #     # print("wcs ra/decs", gfaNum, ra,dec)
        #     gaiaDF = self.getGaiaSources(ra,dec)
        #     gaiaDF["gfaNum"] = gfaNum
        #     allGaia.append(gaiaDF)
        # self.allGaia = pandas.concat(allGaia)
        # self.allGaia = self.allGaia[self.allGaia.gfaNum==1]

        raDecGaia = self.allGaia[["ra", "dec"]].to_numpy()
        # na values in centroids mean no wcs soln was present, so drop them
        cents = self.allCentroids.dropna().reset_index(drop=True)
        # cents = cents[cents.gfaNum==1]
        raDecMeas = cents[["raMeas", "decMeas"]].to_numpy()

        matches, indices, minDists = arg_nearest_neighbor(raDecMeas, raDecGaia)

        gaiaNN = self.allGaia.iloc[indices].reset_index(drop=True)

        matched = pandas.concat([cents, gaiaNN], axis=1)
        matched = matched.loc[:, ~matched.columns.duplicated()].copy()

        # only keep matches closer than specified threshold
        dra = (matched.ra - matched.raMeas)
        dra = dra * numpy.cos(numpy.radians(matched.dec))
        ddec = (matched.dec - matched.decMeas)
        matched["drSky"] = numpy.sqrt(dra**2 + ddec**2) * 3600

        matched = matched[matched.drSky < self.skyDistThresh].reset_index(drop=True)

        # print("nmatched wcs", len(matched))

        # plt.figure(figsize=(8,8))
        # plt.plot(self.allGaia.ra, self.allGaia.dec, 'o', mfc="none", mec="red")
        # plt.plot(matched.raMeas, matched.decMeas, 'x', color="black")
        # plt.savefig("skymatch_%i.png"%self.nIterWCS, dpi=200)
        # plt.close("all")

        # for gfaNum in list(set(matched.gfaNum)):
        #     plt.figure(figsize=(8,8))
        #     ag = self.allGaia[self.allGaia.gfaNum==gfaNum]
        #     mm = matched[matched.gfaNum==gfaNum]
        #     plt.plot(ag.ra, ag.dec, 'o', mfc="none", mec="red")
        #     plt.plot(mm.raMeas, mm.decMeas, 'x', color="black")
        #     plt.savefig("skymatch_%i_gfa%i.png"%(self.nIterWCS, gfaNum), dpi=200)
        #     plt.close("all")

        # import pdb; pdb.set_trace()
        # import time; time.time(); tstart=time.time()
        output = radec2wokxy(
            matched.ra.to_numpy(), matched.dec.to_numpy(), self.GAIA_EPOCH,
            "GFA", self.raCenMeas, self.decCenMeas, self.paCenMeas,
            self.observatory, self.obsTimeRef.jd, focalScale=self.scaleMeas,
            pmra=matched.pmra.to_numpy(), pmdec=matched.pmdec.to_numpy(),
            parallax=matched.parallax.to_numpy(), fullOutput=True
        )
        # print("radec2wokxy wcs took", time.time()-tstart)
        xPred = output[0]
        yPred = output[1]
        fieldWarn = output[2]
        ha = output[3]
        pa = output[4]
        xFocal = output[5]
        yFocal = output[6]
        thetaField = output[7]
        phiField = output[8]
        altPred = output[9]
        azPred = output[10]
        self.altCenMeas = output[11]
        self.azCenMeas = output[12]

        matched["xWokPred"] = xPred
        matched["yWokPred"] = yPred
        matched["dx"] = matched.xWokMeas - matched.xWokPred
        matched["dy"] = matched.yWokMeas - matched.yWokPred
        matched["dr"] = numpy.sqrt(matched.dx**2 + matched.dy**2)

        # plt.figure(figsize=(8,8))
        # plt.plot(matched.xWokPred, matched.yWokPred, 'o', mfc="none", mec="red")
        # plt.plot(matched.xWokMeas, matched.yWokMeas, 'x', color="black")
        # plt.savefig("wok_wcs_%i.png"%self.nIterWCS, dpi=200)
        # plt.close("all")

        # for gfaNum in list(set(matched.gfaNum)):
        #     plt.figure(figsize=(8,8))
        #     mm = matched[matched.gfaNum==gfaNum]
        #     plt.figure(figsize=(8,8))
        #     plt.plot(mm.xWokPred, mm.yWokPred, 'o', mfc="none", mec="red")
        #     plt.plot(mm.xWokMeas, mm.yWokMeas, 'x', color="black")
        #     plt.savefig("wok_wcs_%i_gfa%i.png"%(self.nIterWCS, gfaNum), dpi=200)
        #     plt.close("all")


        self.matchedSources = matched
        self._updateZP()
        # import pdb; pdb.set_trace()
        # print("sources matched", len(self.matchedSources))

    def _matchWok(self):
        # import time; tstart=time.time()
        output = radec2wokxy(
            self.allGaia.ra.to_numpy(), self.allGaia.dec.to_numpy(), self.GAIA_EPOCH,
            "GFA", self.raCenMeas, self.decCenMeas, self.paCenMeas,
            self.observatory, self.obsTimeRef.jd, focalScale=self.scaleMeas,
            pmra=self.allGaia.pmra.to_numpy(), pmdec=self.allGaia.pmdec.to_numpy(),
            parallax=self.allGaia.parallax.to_numpy(), fullOutput=True
        )
        # print("n gaia", len(self.allGaia), "radec2wokxy took", time.time()-tstart)
        xPred = output[0]
        yPred = output[1]
        fieldWarn = output[2]
        ha = output[3]
        pa = output[4]
        xFocal = output[5]
        yFocal = output[6]
        thetaField = output[7]
        phiField = output[8]
        altPred = output[9]
        azPred = output[10]
        self.fieldCenAltMeas = output[11]
        self.fieldCenAzMeas = output[12]

        self.allGaia["xWokPred"] = xPred
        self.allGaia["yWokPred"] = yPred

        xyWokMeas = self.allCentroids[["xWokMeas", "yWokMeas"]].to_numpy()
        xyWokPredict = self.allGaia[["xWokPred", "yWokPred"]].to_numpy()
        # import time; t1 = time.time()
        matches, indices, minDists = arg_nearest_neighbor(xyWokMeas, xyWokPredict)
        # print("arg_nearset took", time.time()-t1)

        gaiaNN = self.allGaia.iloc[indices].reset_index(drop=True)
        matched = pandas.concat([self.allCentroids, gaiaNN], axis=1)
        matched = matched.loc[:, ~matched.columns.duplicated()].copy()

        # reject matches above the threshold
        matched["dx"] = matched.xWokMeas - matched.xWokPred
        matched["dy"] = matched.yWokMeas - matched.yWokPred
        matched["dr"] = numpy.sqrt(matched.dx**2 + matched.dy**2)

        # plt.figure()
        # plt.hist(matched.dr, bins=200)
        # plt.savefig("wokdist.png")
        # plt.close("all")


        # for gfaNum in range(1,7):
        #     g1 = matched[matched.gfaNum==gfaNum]
        #     plt.figure(figsize=(8,8))
        #     plt.plot(g1.xWokPred, g1.yWokPred, 'o', mfc="none", mec="red")
        #     plt.plot(g1.xWokMeas, g1.yWokMeas, 'x', color="black")
        #     plt.savefig("wokmatch_%i_gfa%i.png"%(self.nIterAll, gfaNum), dpi=200)
        #     plt.close("all")

        # plt.figure(figsize=(8,8))
        # plt.plot(matched.xWokPred, matched.yWokPred, 'o', mfc="none", mec="red")
        # plt.plot(matched.xWokMeas, matched.yWokMeas, 'x', color="black")
        # plt.savefig("wokmatch_%i.png"%self.nIterAll, dpi=200)
        # plt.close("all")

        goodMatches = matched[matched.dr < self.wokDistThresh]
        # import pdb; pdb.set_trace()
        self.matchedSources = goodMatches.reset_index(drop=True)
        self._updateZP()

    def _iter(self):
        # matched_df = self.matchedSources
        # print("called _iter()", self.fit_rms)
        nGFAS = len(set(self.matchedSources.gfaNum))

        # x = matched_df.xWokMeas.to_numpy()
        # y = matched_df.yWokMeas.to_numpy()
        # dx = xPred - x
        # dy = yPred - y
        # measRMS = numpy.sqrt(numpy.mean(dx**2 + dy**2)) * 1000
        # print("meas rms", measRMS)

        # plt.figure(figsize=(8,8))
        # plt.quiver(x,y,dx,dy,angles="xy",units="xy", width=.2)
        # plt.axis("equal")
        # plt.show()
        xyMeas = self.matchedSources[["xWokMeas", "yWokMeas"]].to_numpy()
        xyPred = self.matchedSources[["xWokPred", "yWokPred"]].to_numpy()

        if nGFAS > 1:
            st = SimilarityTransform()
            st.estimate(xyPred, xyMeas)
            xyFit = st(xyPred)
            dxy = xyFit - xyMeas
            dRot = st.rotation
            dTrans = st.translation
            # print("dTrans", dTrans, numpy.degrees(dRot), st.scale)
            dScale = st.scale
            self.rotation += dRot
            self.translation += dTrans
        else:
            dTrans = numpy.mean(xyMeas - xyPred, axis=0)
            dxy = xyPred + dTrans - xyMeas
            dRot = 0
            dScale = 1

        # print(st.translation, numpy.degrees(st.rotation), st.scale)

        # self.fit_rms = numpy.sqrt(
        #     numpy.mean(numpy.sum(dxy**2, axis=1))
        # )
        # print("%.8f"%self.fit_rms, len(self.matchedSources), set(self.matchedSources.gfaNum))

        # plt.figure(figsize=(8,8))
        # plt.quiver(x,y,dx,dy,angles="xy", units="xy", width=.2)
        # plt.axis("equal")

        # update field center
        self.paCenMeas += numpy.degrees(dRot)
        rotRad = numpy.radians(self.paCenMeas)
        cosRot = numpy.cos(rotRad)
        sinRot = numpy.sin(rotRad)
        # rotate by PA to get offset directions along ra/dec
        rotMat = numpy.array([
            [cosRot, sinRot],
            [-sinRot, cosRot]
        ])
        dra, ddec = rotMat @ dTrans
        self.decCenMeas -= ddec / self.plate_scale
        dra = dra / self.plate_scale
        dra = dra / numpy.cos(numpy.radians(self.decCenMeas))
        self.raCenMeas -= dra
        self.scaleMeas /= dScale
        # print("len allGaia", len(self.allGaia))

        # plt.show()

    def solve(
        self,
        skyDistThresh: float = 3,  # arcseconds
    ):
        # print("field initial solve\n------------\n")
        self.skyDistThresh = skyDistThresh
        # initialize field center to
        # user supplied reference


        self.nIterWCS = 0
        self.nIterAll = 0
        if len(self.gfaWCS) > 0:
            # match wcs gets gaia sources for matching based on
            # wcs if available
            allGaia = []
            for gfaNum, wcs in self.gfaWCS.items():
                ra, dec = wcs.pixel_to_world_values([[1024,1024]])[0]
                # print("wcs ra/decs", gfaNum, ra,dec)
                gaiaDF = self.getGaiaSources(ra,dec)
                gaiaDF["gfaNum"] = gfaNum
                allGaia.append(gaiaDF)
            self.allGaia = pandas.concat(allGaia)

            lastRMS = None
            self._matchWCS()
            for ii in range(10):
                # maximum of 10 iters
                self._iter()
                self._matchWCS()
                self.nIterWCS += 1
                # print("wcs fit_rms", ii, len(self.matchedSources), self.fit_rms)
                if lastRMS is not None:
                    if numpy.abs(lastRMS - self.fit_rms) < 0.003:
                        break
                lastRMS = self.fit_rms

        if len(self.gfaHeaders) != len(self.gfaWCS):
            # at least one chip missing a WCS,
            # add in gfa's without WCS solution  and re-fit
            gaiaGFAs = list(set(self.allGaia.gfaNum))
            for gfaNum in self.gfaHeaders.keys():
                if gfaNum in gaiaGFAs:
                    continue  # already got gaia sources for this GFA
                ra, dec = self.gfa2radec(gfaNum)
                gaiaDF = self.getGaiaSources(ra, dec)
                gaiaDF["gfaNum"] = gfaNum
                self.allGaia = pandas.concat([self.allGaia, gaiaDF])

            lastRMS = None
            self._matchWok()
            for ii in range(10):
                # print("coordio iter", ii)

                self._iter()
                self._matchWok()
                self.nIterAll += 1
                # print("coordion fit_rms", ii, len(self.matchedSources), self.fit_rms)
                if lastRMS is not None:
                    if numpy.abs(lastRMS - self.fit_rms) < 0.003:
                        break
                lastRMS = self.fit_rms

        return self.guider_fit

    def reSolve(
        self,
        imgNum: int,
        obsTimeRef: Time,
        exptime: float,
        newCentroids: pandas.DataFrame,
        skyDistThresh: float = 3,  # arcseconds
    ):
        """ warning only use if:
                imgNum = self.imgNum + 1 (next image in sequence)
                previous guide rms was small (eg < 1 arcsecond)
                pointing reference is same as previous frame
                same gfas included as previous frame

        centroids needs (at least) columns x,y,peak,gfaNum
        date_obs, exptime come from GFA header

        this skips all db queries and relies on pre-existing
        guide star table and doesn't require any pre-existing WCS
        solutions
        """
        self.skyDistThresh = skyDistThresh
        self.imgNum = imgNum
        self.obsTimeRef = obsTimeRef
        self.exptime = exptime

        # self.saved_nstars = self.n_stars_used
        # self.saved_fit_rms = self.fit_rms
        # self.saved_used_cameras = self.used_cameras

        # print("resolving field")

        dfList = []
        for gfaNum, group in newCentroids.groupby("gfaNum"):
            # group = group[group.peak < 55000].reset_index(drop=True)
            # calculate wok coordinates for each centroid
            xWokMeas, yWokMeas = self.pix2wok(
                group.x.to_numpy(),
                group.y.to_numpy(),
                gfaNum
            )
            group["xWokMeas"] = xWokMeas
            group["yWokMeas"] = yWokMeas
            dfList.append(group)

        self.allCentroids = pandas.concat(dfList).reset_index(drop=True)

        lastRMS = None
        self.nIterAll = 0
        self.nIterWCS = 0

        self._matchWok()
        for ii in range(10):
            # print("coordio iter", ii)
            self._iter()
            self._matchWok()
            self.nIterAll += 1
            # print("coordion fit_rms", ii, len(self.matchedSources), self.fit_rms)
            if lastRMS is not None:
                if numpy.abs(lastRMS - self.fit_rms) < 0.003:
                    break
            lastRMS = self.fit_rms

        return self.guider_fit


            # compute wok position of each centroid
        # return self.result







import numpy
import pandas
import pytest

from coordio import Field, FocalPlane, Observed, Site, Tangent, Wok
from coordio.defaults import (APO_MAX_FIELD_R, LCO_MAX_FIELD_R,
                              MICRONS_PER_MM, calibration)


@pytest.mark.xfail(reason="GFAs not yet handled.")
def test_proj():

    numpy.random.seed(0)

    # from sdssconv.fieldCoords import parallacticAngle
    apoSite = Site("APO")
    lcoSite = Site("LCO")

    # set time (not used but required for Observed coords)
    apoSite.set_time(2458863, scale='TAI')
    lcoSite.set_time(2458863, scale='TAI')

    # start with some field coords and propagate them
    fcAPO = Observed([[80, 120]], site=apoSite)
    fcLCO = Observed([[80, 120]], site=lcoSite)

    nCoords = 200000

    thetaField = numpy.random.uniform(0, 360, size=nCoords)
    phiFieldLCO = numpy.sqrt(numpy.random.uniform(0, LCO_MAX_FIELD_R**2, size=nCoords))
    phiFieldAPO = numpy.sqrt(numpy.random.uniform(0, APO_MAX_FIELD_R**2, size=nCoords))

    fieldAPO = Field(
        numpy.array([thetaField, phiFieldAPO]).T,
        field_center=fcAPO)

    fieldLCO = Field(
        numpy.array([thetaField, phiFieldLCO]).T,
        field_center=fcLCO
    )

    focalAPO = FocalPlane(fieldAPO, site=apoSite)
    focalLCO = FocalPlane(fieldLCO, site=lcoSite)

    obsAngle = 0

    wokAPO = Wok(focalAPO, site=apoSite, obsAngle=obsAngle)
    wokLCO = Wok(focalLCO, site=lcoSite, obsAngle=obsAngle)

    scaleFactor = 1

    # get tangent coords at these locations
    gfaIDs = ["GFA-S1", "GFA-S2", "GFA-S3", "GFA-S4", "GFA-S5", "GFA-S6"]
    robotIDs = [
        "R0C1", "R+13C1", "R+13C14", "R0C27",
        "R-13C14", "R+13C7", "R-13C1"
    ]

    randomIDs = list(numpy.random.choice(calibration.VALID_HOLE_IDS, size=10))

    allIDs = list(set(gfaIDs + robotIDs + randomIDs))

    for wokCoords, site in zip([wokLCO, wokAPO], [lcoSite, apoSite]):
        for holeID in allIDs:
            tc = Tangent(
                wokLCO, site=lcoSite, holeID=holeID,
                scaleFactor=scaleFactor, obsAngle=obsAngle
            )

            df = {
                "x": tc[:, 0],
                "y": tc[:, 1],
                "z": tc[:, 2],
                "r": numpy.sqrt(tc[:, 0]**2 + tc[:, 1]**2),
                "px": tc.xProj,
                "py": tc.yProj,
                "pr": numpy.sqrt(tc.xProj**2 + tc.yProj**2),
                "d": tc.distProj,

            }
            df = pandas.DataFrame(df)
            df = df[df.r < 40] # keep only thinks within 40 mm xy from tangent center

            dx = df.x - df.px
            dy = df.y - df.py
            dispXY = numpy.sqrt(dx**2 + dy**2) * MICRONS_PER_MM
            dispFoc = (df.d - df.z) * MICRONS_PER_MM

            # less than 5 microns diff between direct proj, and actual proj
            assert numpy.max(dispXY) < 7
            assert numpy.max(numpy.abs(dispFoc)) < 1 # micron

            # zffset is no larger than 700 microns from focal surface
            varFoc = (numpy.mean(df.z) - df.z)*MICRONS_PER_MM

            assert numpy.max(numpy.abs(varFoc)) < 300

            # plt.figure()
            # plt.plot(df.r, dispXY, '.')
            # plt.figure()
            # plt.plot(df.r, dispFoc, '.')
            # plt.figure()
            # plt.plot(df.r, df.z*MICRONS_PER_MM, '.')
            # plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-21
# @Filename: test_site.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import astropy.time
import pytest

from coordio import Site, Time, config


def test_site_apo():

    site = Site('APO')

    apo = config['site']['APO']

    assert site.longitude == apo['longitude']
    assert site.latitude == apo['latitude']

    assert site.pressure is not None
    assert site.pressure < 1000.

    assert site.time is None


def test_set_time():

    site = Site('APO')

    # Jan 15th, 2020
    time = Time(2458863.5, scale='TAI')

    site.set_time(time)
    assert site.time == time

    site.set_time(scale='UTC')
    assert site.time != time
    assert site.time.jd == pytest.approx(astropy.time.Time.now().tai.jd,
                                         abs=1e-8)


def test_copy():

    site = Site('APO')
    site_copy = site.copy()

    assert isinstance(site_copy, Site)
    assert site_copy != site


if __name__ == "__main__":
    test_set_time()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-13
# @Filename: calibration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import warnings
from typing import Optional

import pandas

from coordio.exceptions import CoordIOUserWarning


class Calibration:
    """Calibration class.

    This class holds several Pandas data frames containing the wok and
    positioner calibrations. On instantiation it loads the calibration
    from ``$WOKCALIB_DIR``, if it exists. Additional calibrations can
    be manually added by calling `.add_calibration`, which are appended
    to the existing calibrations.

    Parameters
    ----------
    paths : str or or list of str or `None`
        The paths to the directory containing the calibration files.
        Defaults to ``$WOKCALIB_DIR``.

    """

    def __init__(self, paths: Optional[str | list[str]] = None):

        if paths is None:
            paths = os.environ.get('WOKCALIB_DIR', '').split(':')
        self.paths = paths

        self.FP_MODEL = pandas.DataFrame()
        self.wokOrient = pandas.DataFrame()
        self.positionerTable = pandas.DataFrame()
        self.wokCoords = pandas.DataFrame()
        self.fiducialCoords = pandas.DataFrame()
        self.fiberAssignments = pandas.DataFrame()
        self.gfaCoords = pandas.DataFrame()

        self.VALID_HOLE_IDS: set[str] = set([])
        self.VALID_GUIDE_IDS: list[str] = []

        self.sites: list[str] = []

        self.fps_calibs_version = 'unknown'

        for path in self.paths:
            if not os.path.exists(path):
                warnings.warn('Calibration path or $WOKCALIB_DIR are not set. '
                              'Coordinate transformations will fail.',
                              CoordIOUserWarning)
                return

            self.add_calibration(path)

        try:
            import fps_calibrations  # type: ignore
            self.fps_calibs_version = fps_calibrations.get_version()
        except ImportError:
            warnings.warn('Cannot retrieve the version of the wok calibrations. '
                          'Consider adding the root of fps_calibrations to PYTHONPATH.',
                          CoordIOUserWarning)

    def add_calibration(self,
                        path: str,
                        focalPlaneFile='focalPlaneModel.csv',
                        wokOrientFile='wokOrientation.csv',
                        positionerTableFile='positionerTable.csv',
                        wokCoordsFile='wokCoords.csv',
                        fiducialCoordsFile='fiducialCoords.csv',
                        fiberAssignmentsFile='fiberAssignments.csv',
                        gfaCoordsFile="gfaCoords.csv"):
        """Add calibration files from a ``path``."""

        path = str(path)
        if not os.path.exists(path):
            raise FileExistsError(f'Path {path} does not exist.')

        # Focal plane model
        self._add_calibration(path, focalPlaneFile, 'FP_MODEL',
                              indices=['site', 'direction', 'waveCat'])

        # Wok orientation
        self._add_calibration(path, wokOrientFile, 'wokOrient', indices='site')

        # Wok coordinates
        self._add_calibration(path, wokCoordsFile, 'wokCoords',
                              indices=['site', 'holeID'], index_col=0)

        self.VALID_HOLE_IDS = set(self.wokCoords.reset_index().holeID.tolist())
        self.VALID_GUIDE_IDS = [ID for ID in self.VALID_HOLE_IDS if ID.startswith('GFA')]

        # Positioner data
        self._add_calibration(path, positionerTableFile, 'positionerTable',
                              indices=['site', 'holeID'], index_col=0)
        self.sites = list(set(self.positionerTable.index.get_level_values(0)))

        # Fiducials
        self._add_calibration(path, fiducialCoordsFile, 'fiducialCoords',
                              indices=['site', 'holeID'], index_col=0)

        # Fibre assignments
        self._add_calibration(path, fiberAssignmentsFile, 'fiberAssignments',
                              indices=['site', 'holeID'], index_col=0)

        # Fibre assignments
        self._add_calibration(path, gfaCoordsFile, 'gfaCoords',
                              indices=['site', 'id'], index_col=0)

    def _add_calibration(self, path: str, file: str, variable: str,
                         indices: str | list | None = None, index_col=None):
        """Loads and concatenates one file."""

        path = os.path.join(path, file)
        if not os.path.exists(path):
            warnings.warn(f'Cannot find file {path}.', CoordIOUserWarning)
            return

        # Sets the file name.
        setattr(self, variable + 'File', path)

        new = pandas.read_csv(path, comment="#", index_col=index_col)
        if len(new) == 0:
            return path

        # Check that we don't have already loaded a calibration for this site.
        if 'site' not in new:
            warnings.warn(f'Column site not found in {path}', CoordIOUserWarning)
            return
        else:
            new_sites = set(new.site.tolist())

        current = getattr(self, variable)
        if len(current) > 0:
            sites = set(current.reset_index().site.tolist())
            if len(sites & new_sites) > 0:
                raise ValueError('Some new sites already exist in calibration.')

        if indices:
            new.set_index(indices, inplace=True)

        setattr(self, variable, pandas.concat([current, new]))

        return path



import numpy
import pytest


numpy.random.seed(7)
import time

import matplotlib.pyplot as plt
import pandas
import seaborn as sns

from coordio import conv, libcoordio
# from coordio import Site, Wok, Observed, Field, FocalPlane, Tangent
from coordio.defaults import MICRONS_PER_MM, POSITIONER_HEIGHT, calibration


@pytest.mark.xfail()
def test_sameSolution():
    alphaLen = 7.4
    metXY = [14.314, 0.0]
    apXY = [14.965,0.376]
    bossXY = [14.965, -0.376]

    # safe zones
    for theta in numpy.random.uniform(2,350, 20):
        r = numpy.random.uniform(7.5, 22) # mm radius
        alphaOff = numpy.random.uniform(-0.1, 0.1)
        betaOff = numpy.random.uniform(-0.1, 0.1)
        x = r*numpy.cos(numpy.radians(theta))
        y = r*numpy.sin(numpy.radians(theta))

        for fx, fy in [metXY, apXY, bossXY]:
            a1R, b1R, a1L, b1L, err = libcoordio.tangentToPositioner2(
                [x,y], [fx, fy], alphaLen, alphaOff, betaOff
            )

            a2, b2 = libcoordio.tangentToPositioner(
                [x,y], [fx, fy], alphaLen, alphaOff, betaOff
            )

            assert numpy.abs(a1R-a2) < 1e-10
            assert numpy.abs(b1R-b2) < 1e-10
            assert err == 0


def test_outsideDonut():
    alphaLen = 7.4
    metXY = [14.314, 0.0]

    for theta in numpy.random.uniform(2,350, 5):
        r = metXY[0] + alphaLen + 0.01 # mm radius
        alphaOff = 0
        betaOff = 0
        x = r*numpy.cos(numpy.radians(theta))
        y = r*numpy.sin(numpy.radians(theta))

        fx,fy = metXY

        a1R, b1R, a1L, b1L, err = libcoordio.tangentToPositioner2(
            [x,y], [fx, fy], alphaLen, alphaOff, betaOff
        )

        a2, b2 = libcoordio.tangentToPositioner(
            [x,y], [fx, fy], alphaLen, alphaOff, betaOff
        )

        assert numpy.isnan(a2)
        assert numpy.isnan(b2)
        assert b1R == 0
        assert b1L == 0
        assert err > 0

def test_insideDonut():
    alphaLen = 7.4
    metXY = [14.314, 0.0]

    for theta in numpy.random.uniform(2,350, 5):
        r = metXY[0] + alphaLen + 0.01 # mm radius
        alphaOff = 0
        betaOff = 0
        x = r*numpy.cos(numpy.radians(theta))
        y = r*numpy.sin(numpy.radians(theta))

        fx,fy = metXY

        a1R, b1R, a1L, b1L, err = libcoordio.tangentToPositioner2(
            [x,y], [fx, fy], alphaLen, alphaOff, betaOff
        )

        a2, b2 = libcoordio.tangentToPositioner(
            [x,y], [fx, fy], alphaLen, alphaOff, betaOff
        )

        assert numpy.isnan(a2)
        assert numpy.isnan(b2)
        assert b1R == 0
        assert b1L == 0
        assert err > 0
            # print(a1R-a2, b1R-b2)




if __name__ == "__main__":
    test_outsideDonut()



import numpy
import pytest

from coordio import CoordIOError
from coordio.zernike import ZernFit, gradZern, orthoZern


numpy.random.seed(0)

nPts = 1000
badX = numpy.random.uniform(size=nPts)
badY = numpy.random.uniform(size=nPts)

goodX = badX * numpy.sqrt(2) / 2
goodY = badY * numpy.sqrt(2) / 2

zernOrder = 3
largeOrder = 21


def test_ortho():
    dx, dy = orthoZern(goodX, goodY, zernOrder)
    assert dx.shape[0] == nPts
    assert dy.shape[0] == nPts
    assert dx.shape[1] == zernOrder
    assert dy.shape[1] == zernOrder
    with pytest.raises(CoordIOError):
        dx, dy = orthoZern(goodX, goodY, largeOrder)
    with pytest.raises(CoordIOError):
        dx, dy = orthoZern(badX, badY, zernOrder)


def test_grad():
    dx, dy = gradZern(goodX, goodY, zernOrder)
    assert dx.shape[0] == nPts
    assert dy.shape[0] == nPts
    assert dx.shape[1] > zernOrder
    assert dy.shape[1] > zernOrder

    with pytest.raises(CoordIOError):
        dx, dy = gradZern(badX, badY, zernOrder)

    dx, dy = gradZern(goodX, goodY, largeOrder)
    assert dx.shape[0] == nPts
    assert dy.shape[0] == nPts
    assert dx.shape[1] > largeOrder
    assert dy.shape[1] > largeOrder


# Fails sometimes but not always.
@pytest.mark.xfail()
def test_fit():
    zf = ZernFit(goodX, goodY, badX, badY, orders=20, method="ortho")
    xFit, yFit = zf.apply(goodX, goodY)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-12-05 12:01:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-12-05 12:19:32


class CoordIOError(Exception):
    """A custom core coordIO exception"""

    def __init__(self, message=None):

        message = 'There has been an error' \
            if not message else message

        super(CoordIOError, self).__init__(message)


class CoordinateError(CoordIOError):
    """Coordinate error."""


class CoordIOWarning(Warning):
    """Base warning for coordIO."""


class CoordIOUserWarning(UserWarning, CoordIOWarning):
    """The primary warning class."""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-08-22
# @Filename: extraction.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import warnings

import numpy
import pandas
import seaborn
import sep
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.modeling.models import Gaussian1D
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from numpy.typing import NDArray


__all__ = ["sextractor_quick"]


seaborn.set_color_codes("deep")


def sextractor_quick(
    data: NDArray,
    threshold: float = 5.0,
    clean: bool = True,
    minarea: int = 10,
    return_background: bool = False,
) -> pandas.DataFrame | tuple[pandas.DataFrame, NDArray]:
    """Runs a quick extraction using SExtractor (sep).

    Parameters
    ----------
    data
        The array from which to extract sources.
    threshold
        The threshold for detections, in sigmas above the background.
    clean
        Applies some reasonable cuts to select only star-like objects
        (eccentricity greater than 0.05 and less than 0.8).
    minarea
        Minimum number of pixels in the detected regions.
    return_background
        Whether to return the background image.

    Returns
    -------
    extractions
        A data frame with the output of ``sep.extract``. If
        ``return_background=True`` a second element is returned with
        the background image.

    """

    data = data.astype("f8")

    back = sep.Background(data)
    background: NDArray[numpy.float32] = back.back()
    rms = back.globalrms

    stars = sep.extract(data - background, threshold, err=rms, minarea=minarea)

    df = pandas.DataFrame(stars)

    df["ecc"] = numpy.sqrt(df.a**2 - df.b**2) / df.a

    if clean:
        filter = (df.cpeak < 60000 - back.globalback) & (df.ecc < 0.8) & (df.ecc > 0.05)
        df = df.loc[filter]

    df = df.loc[df.tnpix > minarea]

    if return_background:
        return df, background

    return df


def get_marginal(
    data: NDArray,
    x: int,
    y: int,
    box_size: int,
    axis: int = 0,
    normalise: bool | float = True,
):
    """Returns the marginal distribution around a point.

    Parameters
    ----------
    data
        The image array.
    x
        The x coordinate of the centre of the box. If not an integer, it's
        truncated.
    y
        The y coordinate of the centre of the box.
    box_size
        The size of the box around the ``x,y`` coordinates. Either an
        odd number or the closes odd number will be used.
    axis
        The axis of ``data`` along which to collapse the image.
    normalise
        Whether to normalise the marginal distribution. If a number, the
        amount by which to normalise.

    Returns
    -------
    marginal
        The normalised marginal distribution of the box around ``x,y``
        summed along axis ``axis``.

    """

    box_size = int(box_size)
    if box_size % 2 == 0:
        box_size += 1

    x = int(x)
    y = int(y)

    marginal = data[
        y - box_size // 2 : y + box_size // 2 + 1,
        x - box_size // 2 : x + box_size // 2 + 1,
    ]
    marginal = marginal.sum(axis=axis)

    if normalise is not False:
        if normalise is True:
            marginal /= marginal[box_size // 2 - 3 : box_size // 2 + 3].max()
        else:
            marginal /= normalise

    return marginal


def fit_gaussian_to_marginal(
    data: NDArray,
    x: int,
    y: int,
    box_size: int,
    axis: int = 0,
    peak: float | None = None,
    sigma_0: float | None = None,
):
    """Fits the marginal distribution with a 1D Gaussian.

    Parameters
    ----------
    data
        The image array.
    x
        The x coordinate of the centre of the box. If not an integer, it's
        truncated.
    y
        The y coordinate of the centre of the box.
    box_size
        The size of the box around the ``x,y`` coordinates. Either an
        odd number or the closes odd number will be used.
    axis
        The axis of ``data`` along which to collapse the image.
    peak
        The peak of the detection. If passed, used for normalisation.
    sigma_0
        An initial estimate of the sigma of the Gaussian to fit.

    Returns
    -------
    fit
        A tuple containing the mean of the fitted Gaussian, the fitted sigma,
        and the RMS of the residuals between the original data and the fit.

    """

    box_size = int(box_size)
    if box_size % 2 == 0:
        box_size += 1

    x = int(x)
    y = int(y)

    marginal = get_marginal(
        data,
        x,
        y,
        box_size,
        axis=axis,
        normalise=True if peak is None else peak,
    )
    xx = numpy.arange(marginal.size)

    fitter = LevMarLSQFitter()
    model = Gaussian1D(2, box_size / 2, sigma_0)
    model.amplitude.bounds = (0.1, 1.1)

    valid = 1
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        gg = fitter(model, xx, marginal)

        if fitter.fit_info["cov_x"] is None:
            valid = 0

    rms = numpy.sqrt(numpy.sum((marginal - gg(xx)) ** 2) / (len(xx) - 1))

    if axis == 0:
        mean = y + gg.mean - box_size // 2
    else:
        mean = x + gg.mean - box_size // 2

    return (mean, gg.stddev.value, rms, valid)


def extract_marginal(
    data: NDArray,
    threshold: float = 5.0,
    box_size: int = 51,
    sigma_0: float | None = None,
    max_detections: int | None = None,
    exclude_border: bool = True,
    sextractor_quick_options: dict = {},
    plot: pathlib.Path | str | None = None,
    n_rows_plot: int = 9,
    plot_title: str | None = None,
    aper_radius: float | None = None
):
    """Extracts regions from an image and fit them using the marginal distribution.

    Parameters
    ----------
    data
        The image array.
    threshold
        The threshold for detections, in sigmas above the background.
    box_size
        The size of the box around the ``x,y`` coordinates. Either an
        odd number or the closes odd number will be used.
    sigma_0
        An initial estimate of the sigma of the Gaussian to fit.
    max_detections
        If passed, the maximum number of detections to fit. If the number
        of SExtractor detections is larger than ``max_detections`` the top
        ones, sorted by flux, will be used.
    exclude_border
        Reject regions that are closer to the border of the image than
        ``box_size``.
    sextractor_quick_options
        Additional parameters to pass to `.sectractor_quick`.
    plot
        If `None`, no plots will be produced. Otherwise it must
        be the path where to write the PDF with the plot.
    n_rows_plot
        How many regions per page to plot.
    plot_title
        A title to be added to the plot.
    aper_radius
        radius in pixels for aperture photometry


    Returns
    -------
    extractions
        A data frame with the extracted regions. It includes all the
        usual columns from ``sep.extract`` plus ``x1``, ``xstd``, ``xrms``
        (and same for ``y``) with the mean, sigma, and RMS for the given
        axis obtained by fitting the marginal distribution (see
        `.fit_gaussian_to_marginal`).

    """

    # print("extractor_quick_options", sextractor_quick_options)
    data = data.astype("f8")

    sextractor_quick_options.pop("threshold", None)
    sextractor_quick_options.pop("return_background", None)

    detections, back = sextractor_quick(
        data,
        threshold=threshold,
        return_background=True,
        **sextractor_quick_options,
    )

    assert isinstance(detections, pandas.DataFrame)
    assert isinstance(back, numpy.ndarray)

    if len(detections) == 0:
        return detections

    if exclude_border:
        detections = detections.loc[
            (detections.x > box_size // 2)
            & (detections.x < data.shape[1] - box_size // 2)
            & (detections.y > box_size // 2)
            & (detections.y < data.shape[0] - box_size // 2),
            :,
        ]

    detections.sort_values("flux", ascending=False, inplace=True)
    if max_detections and len(detections) > max_detections:
        detections = detections.head(max_detections)

    back = sep.Background(data)
    sub = data - back.back()

    if len(detections) > 0:
        for axis in [1, 0]:
            ax = "x" if axis == 1 else "y"

            fit_df = detections.apply(
                lambda d: pandas.Series(
                    fit_gaussian_to_marginal(
                        sub,
                        d.x,
                        d.y,
                        box_size,
                        axis=axis,
                        sigma_0=sigma_0,
                    ),
                    index=[f"{ax}1", f"{ax}std", f"{ax}rms", f"{ax}fitvalid"],
                ),
                axis=1,
            )

            detections = pandas.concat([detections, fit_df], axis=1)

        # circular aperture fluxes (5 arcsec radius, as specified by
        # tom dwelly)
        # could use x1 or y1 instead?
        xs = detections.x.to_numpy()
        ys = detections.y.to_numpy()
        if aper_radius is None:
            # if not specified set the radius
            # to the FWHM
            _stds = numpy.array(
                [detections.xstd.to_numpy(),detections.ystd.to_numpy()]
            ).flatten()
            # https://brainder.org/2011/08/20/gaussian-kernels-convert-fwhm-to-sigma/
            _fwhm = numpy.median(_stds) * 2.355
            # print("_fwhm", _fwhm)
            aper_radius = _fwhm

        flux, fluxerr, flag = sep.sum_circle(
            sub, xs, ys, aper_radius, err=back.globalrms
        )

        detections["aperflux"] = flux
        detections["aperfluxerr"] = fluxerr
        detections["aperrad"] = aper_radius

        # detections["aperflux"] = detections.flux
        # import pdb; pdb.set_trace()

    else:
        # Add new columns. If there are no detections at least the columns will exist
        # on an empty data frame and the overall shape won't change.
        cols = ["x1", "xstd", "xrms", "y1", "ystd", "yrms", "xfitvalid", "yfitvalid"]
        cols += ["aperflux", "aperfluxerr", "aperrad"]
        detections[cols] = numpy.nan

    if plot is not None:
        if not isinstance(plot, pathlib.Path) and not isinstance(plot, str):
            raise TypeError("plot must be the path to the output file.")

        plot = pathlib.Path(plot)
        plot.parent.mkdir(exist_ok=True, parents=True)

        with plt.ioff():
            with PdfPages(str(plot)) as pdf:

                figure, ax = plt.subplots(figsize=(8.5, 8.5))
                ax.imshow(
                    sub,
                    origin="lower",
                    vmin=sub.mean() - sub.std(),
                    vmax=sub.mean() + sub.std(),
                    cmap="gray",
                )

                if len(detections) > 0:
                    ax.scatter(
                        detections.x,
                        detections.y,
                        marker="x",  # type:ignore
                        c="r",
                    )

                    for reg, row in detections.iterrows():
                        ax.annotate(
                            str(reg),
                            (row.x + 15, row.y + 15),
                            fontsize=10,
                            color="k",
                        )

                ax.set_title(plot_title or "", fontsize=15, pad=20)

                pdf.savefig(figure)
                plt.close(figure)

                for ii in range(0, len(detections), n_rows_plot):
                    sample = detections.iloc[ii : ii + n_rows_plot]
                    if len(sample) == 0:
                        break
                    _plot_one_page(pdf, sub, sample, box_size, n_rows_plot)

            plt.close("all")

            plt.clf()
            plt.cla()

    return detections


def _plot_one_page(
    pdf: PdfPages,
    data: NDArray,
    df: pandas.DataFrame,
    box_size: int,
    n_rows: int,
):
    """Plots one page of marginal distributions."""

    figure, ax = plt.subplots(n_rows, 2, figsize=(8.5, 11))

    ii = 0
    for region, row in df.iterrows():

        for col_ax, axis in enumerate(["x", "y"]):

            marginal = get_marginal(
                data,
                row.x,
                row.y,
                box_size,
                axis=0 if axis == "y" else 1,
            )

            xx = numpy.arange(marginal.size) + row[axis] - box_size // 2
            ax[ii][col_ax].plot(
                xx,
                marginal,
                color="k",
                ls="dotted",
            )

            if axis == "x":
                mean = row.x1
                rms = row.xrms
                stddev = row.xstd
            else:
                mean = row.y1
                rms = row.yrms
                stddev = row.ystd

            model = Gaussian1D(1, mean, stddev)
            ax[ii][col_ax].plot(
                xx,
                model(xx),
                color="r",
                ls="solid",
            )

            title = ""
            if axis == "x":
                title += f"Region {region}: "
            title += rf"$\overline{{{axis}}}={mean:.1f},\ $"
            title += rf"$\sigma_{axis}={stddev:.2f},\ $"
            title += f"RMS={rms:.3f}"

            ax[ii][col_ax].set_title(title, fontsize=10)
            ax[ii][col_ax].tick_params(labelsize=10)
            ax[ii][col_ax].tick_params(labelsize=10)

        ii += 1

    if ii < n_rows - 1:
        for jj in range(ii, n_rows):
            ax[jj][0].axis("off")
            ax[jj][1].axis("off")

    figure.subplots_adjust(hspace=0.6, top=0.97, bottom=0.03, left=0.05, right=0.95)

    pdf.savefig(figure)
    plt.close(figure)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-17
# @Filename: sky.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# IAU-defined sky coordinate systems and transformations.

from __future__ import annotations

import ctypes
import warnings
from typing import TYPE_CHECKING

import numpy

from . import conv, defaults, sofa
from .coordinate import Coordinate, Coordinate2D, verifySite, verifyWavelength
from .exceptions import CoordIOError, CoordIOUserWarning
from .time import Time

if TYPE_CHECKING:
    from .site import Site


__all__ = ['ICRS', 'Observed']


class ICRS(Coordinate2D):
    """A representation of ICRS coordinates.

    Parameters
    ----------
    value : numpy.ndarray
        A Nx2 Numpy array with the RA and Dec coordinates of the targets.
    epoch : numpy.ndarray
        A 1D array with the epoch of the coordinates for each target,
        as a TDB Julian date (although for most applications the small
        differences between scales will not matter). Defaults to J2000.
    pmra : numpy.ndarray
        A 1D array with the proper motion in the RA axis for the N targets,
        in milliarcsec/yr. Must be a true angle, i.e, it must include the
        ``cos(dec)`` term.
    pmdec : numpy.ndarray
        A 1D array with the proper motion in the RA axis for the N targets,
        in milliarcsec/yr.
    parallax : numpy.ndarray
        A 1D array with the parallax for the N targets, in milliarcsec.
    rvel : numpy.ndarray
        A 1D array with the radial velocity in km/s, positive when receding.

    """

    __extra_arrays__ = ['epoch', 'pmra', 'pmdec', 'parallax', 'rvel']

    def __new__(cls, value, **kwargs):

        obj = super().__new__(cls, value, **kwargs)

        if kwargs.get('epoch', None) is None:
            obj.epoch += defaults.EPOCH

        if isinstance(value, Coordinate):

            if value.coordSysName == 'Observed':
                obj._fromObserved(value)

            else:
                raise CoordIOError(
                    'Cannot convert to ICRS from %s' % value.coordSysName
                )

        return obj

    def _fromObserved(self, obsCoords):
        """Converts from `.Observed` coordinates.  Epoch is the
        time specifified by the site.

        """

        rlong = numpy.radians(obsCoords.site.longitude)
        rlat = numpy.radians(obsCoords.site.latitude)
        rZD = numpy.radians(90 - obsCoords[:, 0])
        rAz = numpy.radians(obsCoords[:, 1])
        wavelength = obsCoords.wavelength / 10000.

        _type = "A".encode()  # coords are azimuth, zenith dist

        time = obsCoords.site.time

        utc = time.to_utc()
        utc1 = int(utc)
        utc2 = utc - utc1
        dut1 = time.get_dut1()

        _ra = ctypes.c_double()
        _dec = ctypes.c_double()

        ra = numpy.zeros(len(obsCoords))
        dec = numpy.zeros(len(obsCoords))

        for ii in range(len(rAz)):

            sofa.iauAtoc13(
                _type, rAz[ii], rZD[ii], utc1, utc2, dut1,
                rlong, rlat, obsCoords.site.altitude, 0.0, 0.0,
                obsCoords.site.pressure, obsCoords.site.temperature,
                obsCoords.site.rh, wavelength[ii], _ra, _dec
            )
            ra[ii] = numpy.degrees(_ra.value)
            dec[ii] = numpy.degrees(_dec.value)

        self[:, 0] = ra
        self[:, 1] = dec

    def to_epoch(self, jd, site=None):
        """Convert the coordinates to a new epoch.

        Parameters
        ----------
        jd : float
            The Julian date, in TAI scale, of the output epoch.
        site : .Site
            The site of the observation. Used to determine the TDB-TT offset.
            If not provided, it assumes longitude and latitude zero.

        Returns
        -------
        icrs : `.ICRS`
            A new `.ICRS` object with the coordinates, proper motion, etc. in
            the new epoch.

        """

        rra = numpy.radians(self[:, 0])
        rdec = numpy.radians(self[:, 1])

        rpmra = numpy.radians(self.pmra / 1000. / 3600.) / numpy.cos(rdec)
        rpmdec = numpy.radians(self.pmdec / 1000. / 3600.)

        # Using TDB is probably an overkill.

        tai = Time(jd, scale='TAI')

        if site:
            epoch2 = tai.to_tdb(longitude=site.longitude,
                                latitude=site.latitude,
                                altitude=site.altitude)
        else:
            epoch2 = tai.to_tdb()

        epoch2_1 = int(epoch2)
        epoch2_2 = epoch2 - epoch2_1

        ra2 = ctypes.c_double()
        dec2 = ctypes.c_double()
        pmra2 = ctypes.c_double()
        pmdec2 = ctypes.c_double()
        parallax2 = ctypes.c_double()
        rvel2 = ctypes.c_double()

        new_icrs = ICRS(numpy.zeros(self.shape, dtype=self.dtype))

        for ii in range(self.shape[0]):

            epoch1_1 = float(int(self.epoch[ii]))
            epoch1_2 = self.epoch[ii] - epoch1_1

            res = sofa.iauPmsafe(rra[ii], rdec[ii], rpmra[ii], rpmdec[ii],
                                 self.parallax[ii] / 1000., self.rvel[ii],
                                 epoch1_1, epoch1_2, epoch2_1, epoch2_2,
                                 ra2, dec2, pmra2, pmdec2, parallax2, rvel2)

            if res > 1 or res < 0:
                warnings.warn(f'iauPmsafe return with error code {res}.',
                              CoordIOUserWarning)

            new_icrs[ii, :] = numpy.rad2deg([ra2.value, dec2.value])
            new_icrs.pmra[ii] = numpy.rad2deg(pmra2.value) * 3600. * 1000.
            new_icrs.pmra[ii] *= numpy.cos(dec2.value)
            new_icrs.pmdec[ii] = numpy.rad2deg(pmdec2.value) * 3600. * 1000.
            new_icrs.parallax[ii] = parallax2.value * 1000.
            new_icrs.rvel[ii] = rvel2.value
            new_icrs.epoch[ii] = jd

        return new_icrs


class Observed(Coordinate2D):
    """The observed coordinates of a series of targets.

    The array contains the Alt/Az coordinates of the targets. Their RA/Dec
    coordinates can be accessed via the ``ra`` and ``dec`` attributes.
    If `.ICRS` or `.Field` is passed, Alt/Az coordinates are computed.

    Parameters
    ----------
    value : numpy.ndarray
        A Nx2 Numpy array with the Alt and Az coordinates of the targets,
        in degrees.  Or `.ICRS` instance.  Or a `.Field` instance.
    wavelength : numpy.ndarray
        A 1D array with he observing wavelength, in angstrom.
        If not explicitly passed, it tries to inheret from value.wavelength,
        if that doesn't exist, it is set to default specified in:
        `defaults.wavelength`
    site : .Site
        The site from where observations will occur, along with the time
        of the observation.  Mandatory argument.

    Attributes
    -----------
    ra : numpy.ndarray
        Nx1 Numpy array, observed RA in degrees
    dec : numpy.ndarray
        Nx1 Numpy array, observed Dec in degrees
    ha : numpy.ndarray
        Nx1 Numpy array, hour angle in degrees
    pa : numpy.ndarray
        Nx1 Numpy array, position angle in degrees.  By SOFA: the angle between
        the direction to the north celestial pole and direction to the zenith.
        range is [-180, 180].  The sign is according to:
        -ha --> -pa, +ha --> +pa

    """

    __extra_arrays__ = ['wavelength']
    __extra_params__ = ['site']  # mandatory
    __computed_arrays__ = ['ra', 'dec', 'ha', 'pa']

    ra: numpy.ndarray
    dec: numpy.ndarray
    ha: numpy.ndarray
    pa: numpy.ndarray

    wavelength: numpy.ndarray
    site: Site

    def __new__(cls, value, **kwargs):
        # should we do range checks (eg alt < 90)? probably.

        verifySite(kwargs)
        verifyWavelength(kwargs, len(value), strict=False)

        obj = super().__new__(cls, value, **kwargs)

        # check if a coordinate was passed that we can just
        # 'cast' into Observed
        if isinstance(value, Coordinate):

            if value.coordSysName == 'ICRS':
                obj._fromICRS(value)

            elif value.coordSysName == 'Field':
                obj._fromField(value)

            else:
                raise CoordIOError(
                    'Cannot convert to Observed from %s' % value.coordSysName
                )

        else:
            # raw numpy array supplied compute values
            obj._fromRaw()

        return obj

    def _fromICRS(self, icrsCoords):
        """Converts from ICRS to topocentric observed coordinates for a site.
        Automatically executed after initialization with `.ICRS`.

        Computes and sets ra, dec, ha, pa arrays.

        Parameters:
        ------------
        icrsCoords : `.ICRS`
            ICRS coordinates from which to convert to observed coordinates

        """

        # eventually move this to coordio.conv?

        # Prepare to call iauAtco13
        #  Given:
        #     rc,dc  double   ICRS right ascension at J2000.0 (radians)
        #     pr     double   RA proper motion (radians/year)
        #     pd     double   Dec proper motion (radians/year)
        #     px     double   parallax (arcsec)
        #     rv     double   radial velocity (km/s, +ve if receding)
        #     utc1   double   UTC as a 2-part...
        #     utc2   double   ...quasi Julian Date
        #     dut1   double   UT1-UTC (seconds)
        #     elong  double   longitude (radians, east +ve)
        #     phi    double   latitude (geodetic, radians)
        #     hm     double   height above ellipsoid (m, geodetic)
        #     xp,yp  double   polar motion coordinates (radians)
        #     phpa   double   pressure at the observer (hPa = mB)
        #     tc     double   ambient temperature at the observer (deg C)
        #     rh     double   relative humidity at the observer (range 0-1)
        #     wl     double   wavelength (micrometers)
        #
        #  Returned:
        #     aob    double*  observed azimuth (radians: N=0,E=90)
        #     zob    double*  observed zenith distance (radians)
        #     hob    double*  observed hour angle (radians)
        #     dob    double*  observed declination (radians)
        #     rob    double*  observed right ascension (CIO-based, radians)
        #     eo     double*  equation of the origins (ERA-GST)

        # TODO: maybe write this as Cython or C?

        # We need the epoch to be J2000.0 because that's what iauAtco13 likes.
        icrs_2000 = icrsCoords.to_epoch(2451545.0, site=self.site)

        rra = numpy.radians(icrs_2000[:, 0])
        rdec = numpy.radians(icrs_2000[:, 1])
        rpmra = numpy.radians(icrs_2000.pmra / 1000. / 3600.) / numpy.cos(rdec)
        rpmdec = numpy.radians(icrs_2000.pmdec / 1000. / 3600.)

        rlong = numpy.radians(self.site.longitude)
        rlat = numpy.radians(self.site.latitude)

        time = self.site.time

        utc = time.to_utc()
        utc1 = int(utc)
        utc2 = utc - utc1
        dut1 = time.get_dut1()

        az_obs = ctypes.c_double()
        zen_obs = ctypes.c_double()
        ha_obs = ctypes.c_double()
        dec_obs = ctypes.c_double()
        ra_obs = ctypes.c_double()
        eo_obs = ctypes.c_double()

        for ii in range(len(rra)):

            sofa.iauAtco13(
                rra[ii], rdec[ii], rpmra[ii], rpmdec[ii],
                icrs_2000.parallax[ii] / 1000., icrs_2000.rvel[ii],
                utc1, utc2, dut1,
                rlong, rlat, self.site.altitude, 0.0, 0.0,
                self.site.pressure, self.site.temperature,
                self.site.rh, self.wavelength[ii] / 10000.,
                az_obs, zen_obs, ha_obs, dec_obs, ra_obs, eo_obs
            )

            altAz = [
                90 - numpy.rad2deg(zen_obs.value),
                numpy.rad2deg(az_obs.value)
            ]
            self[ii, :] = altAz

            self.ra[ii] = numpy.rad2deg(ra_obs.value)
            self.dec[ii] = numpy.rad2deg(dec_obs.value)
            self.ha[ii] = numpy.rad2deg(ha_obs.value)

            # compute the pa
            self.pa[ii] = numpy.rad2deg(
                sofa.iauHd2pa(ha_obs.value, dec_obs.value, rlat)
            )

    def _fromField(self, fieldCoords):
        """Converts from field coordinates to topocentric observed
        coordinates for a site. Automatically executed after initialization
        with `.Field`.

        Computes and sets ra, dec, ha, pa arrays.

        Parameters:
        ------------
        fieldCoords : `.Field`
            Field coordinates from which to convert to observed coordinates

        """
        # get field center info
        altCenter, azCenter = fieldCoords.field_center.flatten()
        pa = float(fieldCoords.field_center.pa)  # parallactic angle

        alt, az = conv.fieldToObserved(
            fieldCoords.x, fieldCoords.y, fieldCoords.z,
            altCenter, azCenter, pa
        )

        self[:, 0] = alt
        self[:, 1] = az

        self._fromRaw()

    def _fromRaw(self):
        """Automatically executed after initialization with
        an Nx2 `numpy.ndarray` of Alt/Az coords.

        Computes and sets ra, dec, ha, pa arrays.

        """

        self[:, 1] = self[:, 1] % 360

        # compute ra, dec, ha, pa here...
        dec_obs = ctypes.c_double()
        ha_obs = ctypes.c_double()
        rlat = numpy.radians(self.site.latitude)
        rlong = numpy.radians(self.site.longitude)
        ut1 = self.site.time.to_ut1()

        for ii, (alt, az) in enumerate(self):
            raz = numpy.radians(az)
            ralt = numpy.radians(alt)
            sofa.iauAe2hd(raz, ralt, rlat, ha_obs, dec_obs)
            self.ha[ii] = numpy.degrees(ha_obs.value)
            self.dec[ii] = numpy.degrees(dec_obs.value)
            self.pa[ii] = numpy.degrees(
                sofa.iauHd2pa(ha_obs.value, dec_obs.value, rlat)
            )
            # earth rotation angle (from SOFA docs)
            # https://www.iausofa.org/2017_0420_C/sofa/sofa_ast_c.pdf
            era = sofa.iauEra00(ut1, 0)  # time is sum of the 2 args
            _ra = numpy.degrees(era + rlong - ha_obs.value)
            _ra = _ra % 360  # wrap ra

            self.ra[ii] = _ra

    @classmethod
    def fromEquatorial(cls, coords, hadec=False, site=None, **kwargs):
        """Initialises `.Observed` coordinates  from equatorial topocentric.

        Parameters
        ----------
        coords : numpy.ndarray
            Nx2 Numpy array with the RA and Dec coordinates of the targets.
            These must be topocentric equatorial coordinates that will be
            converted to horizontal topocentric used to initialise `.Observed`.
            Alternatively, if ``hadec=True`` then the first column in the array
            must be the hour angle in degrees.
        hadec : bool
            Whether the coordinates are HA and declination.
        site : .Site
            The site from where observations will occur, along with the time
            of the observation.  Mandatory argument.
        kwargs
            Other arguments to pass to `.Observed`.

        """

        coords = numpy.array(coords)

        if len(coords.shape) != 2 or coords.shape[1] != 2:
            raise CoordIOError('coords must be Nx2 array.')

        if site is None:
            raise CoordIOError('A site must be specified.')

        lat_r = numpy.radians(site.latitude)
        dec_d = coords[:, 1]
        dec_r = numpy.radians(dec_d)

        if hadec is False:
            ra_d = coords[:, 0]
            ra_r = numpy.radians(ra_d)

            ut = site.time.to_ut1()
            tt = site.time.to_tt()

            gmst = sofa.iauGmst00(ut, 0.0, tt, 0.0)
            lst_r = gmst + numpy.radians(site.longitude)

            ha_r = lst_r - ra_r

        else:
            ha_r = numpy.radians(coords[:, 0])

        # h = altitude, A = azimuth
        sin_h = (numpy.sin(dec_r) * numpy.sin(lat_r) +
                 numpy.cos(dec_r) * numpy.cos(lat_r) * numpy.cos(ha_r))
        h_r = numpy.arcsin(sin_h)
        h_d = numpy.degrees(h_r)

        sin_A = -numpy.sin(ha_r) * numpy.cos(dec_r) / numpy.cos(h_r)
        cos_A = ((numpy.sin(dec_r) - numpy.sin(lat_r) * numpy.sin(h_r)) /
                 (numpy.cos(lat_r) * numpy.cos(h_r)))
        A_r = numpy.arctan2(sin_A, cos_A)
        A_d = numpy.degrees(A_r) % 360

        return cls(numpy.array([h_d, A_d]).T, site=site, **kwargs)



from coordio.utils import object_offset, Moffat2dInterp, offset_definition
import numpy
import pytest


mag_limits = {}
mag_limits['bright'] = {}
mag_limits['bright']['Boss'] = numpy.array([12.7, -999.0, 12.7, -999.0, 13.0, 13.0, 13.0, -999.0, -999.0, -999.0])
mag_limits['bright']['Apogee'] = numpy.array([-999.0, -999.0, -999.0, -999.0, -999.0, -999.0, -999.0, -999.0, 7.0, -999.0])

mag_limits['dark'] = {}
mag_limits['dark']['Boss'] = numpy.array([15.0, 15.0, 15.0, -999.0, -999.0, -999.0, -999.0, -999.0, -999.0, -999.0])
mag_limits['dark']['Apogee'] = numpy.array([-999.0, -999.0, -999.0, -999.0, -999.0, -999.0, -999.0, -999.0, 7.0, -999.0])

fmagloss = Moffat2dInterp()


def test_all_flags():
    def  test_flags(flag, mag, mag_limits, lunation, waveName,
                    sky, offset_min_skybrightness, can_off):
        # test APO fails with 1D array
        with pytest.raises(ValueError, match='mags must be a 2D numpy.array of shape \\(N, 10\\)'):
            delta_ra, delta_dec, offset_flag = object_offset(mag, mag_limits, lunation,
                                                             waveName, 'APO', fmagloss=fmagloss,
                                                             skybrightness=sky,
                                                             offset_min_skybrightness=offset_min_skybrightness,
                                                             can_offset=can_off)

        # test LCO fails with 1D array
        with pytest.raises(ValueError, match='mags must be a 2D numpy.array of shape \\(N, 10\\)'):
            delta_ra, delta_dec, offset_flag = object_offset(mag, mag_limits, lunation,
                                                             waveName, 'LCO', fmagloss=fmagloss,
                                                             skybrightness=sky,
                                                             offset_min_skybrightness=offset_min_skybrightness,
                                                             can_offset=can_off)

        if can_off is None:
            can_off_arr = None
        else:
            can_off_arr = numpy.array([can_off, can_off])
        # test APO with 2D array
        delta_ra, delta_dec, offset_flag = object_offset(numpy.vstack((mag, mag)), mag_limits, lunation,
                                                         waveName, 'APO', fmagloss=fmagloss, skybrightness=sky,
                                                         offset_min_skybrightness=offset_min_skybrightness,
                                                         can_offset=can_off_arr)
        assert numpy.all(offset_flag == flag)
        assert numpy.all(delta_dec == 0.)
        if flag > 0:
            assert numpy.all(delta_ra == 0.)
        else:
            assert numpy.all(delta_ra > 0.)

        # test LCO with 2D array
        delta_ra, delta_dec, offset_flag = object_offset(numpy.vstack((mag, mag)), mag_limits, lunation,
                                                         waveName, 'LCO', fmagloss=fmagloss, skybrightness=sky,
                                                         offset_min_skybrightness=offset_min_skybrightness,
                                                         can_offset=can_off_arr)
        assert numpy.all(offset_flag == flag)
        assert numpy.all(delta_dec == 0.)
        if flag > 0:
            assert numpy.all(delta_ra == 0.)
        else:
            assert numpy.all(delta_ra > 0.)
    # Boss Bright
    offset_min_skybrightness = 1
    waveName = 'Boss'
    lunation = 'bright'
    flags_test = [0, 1, 2, 8, 16, 32]
    flags_test = [f if f == 0 else f + 64 for f in flags_test]

    test_mags = []
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m + 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([-999.] * len(mag_limits[lunation][waveName])))
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([5. if m != -999. else m for m in mag_limits[lunation][waveName]]))

    skybrightness = [None, None, None, 0.3, None, None]
    can_offset = [None, None, None, None, False, None]
    for flag, mag, sky, can_off in zip(flags_test, test_mags, skybrightness, can_offset):
        test_flags(flag, mag, mag_limits[lunation][waveName], lunation, waveName,
                   sky, offset_min_skybrightness, can_off)

    # Boss dark
    offset_min_skybrightness = 1
    waveName = 'Boss'
    lunation = 'dark'
    flags_test = [0, 1, 2, 8, 16, 32]
    flags_test = [f if f == 0 else f + 64 for f in flags_test]
    
    test_mags = []
    test_mags.append(numpy.array([m - 1 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m + 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([-999.] * len(mag_limits[lunation][waveName])))
    test_mags.append(numpy.array([m - 1 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m - 1 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([12. if m != -999. else m for m in mag_limits[lunation][waveName]]))

    skybrightness = [None, None, None, 0.3, None, None]
    can_offset = [None, None, None, None, False, None]
    for flag, mag, sky, can_off in zip(flags_test, test_mags, skybrightness, can_offset):
        test_flags(flag, mag, mag_limits[lunation][waveName], lunation, waveName,
                   sky, offset_min_skybrightness, can_off)

    # Apogee bright
    offset_min_skybrightness = 1
    waveName = 'Apogee'
    lunation = 'bright'
    flags_test = [0, 1, 2, 8, 16, 32]
    flags_test = [f if f == 0 else f + 64 for f in flags_test]
    
    test_mags = []
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m + 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([-999.] * len(mag_limits[lunation][waveName])))
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([0.1 if m != -999. else m for m in mag_limits[lunation][waveName]]))

    skybrightness = [None, None, None, 0.3, None, None]
    can_offset = [None, None, None, None, False, None]
    for flag, mag, sky, can_off in zip(flags_test, test_mags, skybrightness, can_offset):
        test_flags(flag, mag, mag_limits[lunation][waveName], lunation, waveName,
                   sky, offset_min_skybrightness, can_off)

    # Apogee dark
    offset_min_skybrightness = 1
    waveName = 'Apogee'
    lunation = 'dark'
    flags_test = [0, 1, 2, 8, 16, 32]
    flags_test = [f if f == 0 else f + 64 for f in flags_test]
    
    test_mags = []
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m + 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([-999.] * len(mag_limits[lunation][waveName])))
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags.append(numpy.array([0.1 if m != -999. else m for m in mag_limits[lunation][waveName]]))

    skybrightness = [None, None, None, 0.3, None, None]
    can_offset = [None, None, None, None, False, None]
    for flag, mag, sky, can_off in zip(flags_test, test_mags, skybrightness, can_offset):
        test_flags(flag, mag, mag_limits[lunation][waveName], lunation, waveName,
                   sky, offset_min_skybrightness, can_off)

    # test engineering design_mode
    # Boss Bright
    offset_min_skybrightness = 1
    waveName = 'Boss'
    lunation = 'bright'
    flags_test = [64]

    test_mags = []
    test_mags.append(numpy.array([m - 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))

    skybrightness = [None]
    can_offset = [None]
    for flag, mag, sky, can_off in zip(flags_test, test_mags, skybrightness, can_offset):
        test_flags(flag, mag, numpy.zeros(10) - 999., lunation, waveName,
                   sky, offset_min_skybrightness, can_off)

    # test get combination of all flags
    # Boss Bright
    offset_min_skybrightness = 1
    waveName = 'Boss'
    lunation = 'bright'
    flags_test = [1 + 32]
    flags_test = [f if f == 0 else f + 64 for f in flags_test]

    test_mags = []
    test_mags.append(numpy.array([m + 2 if m != -999. else m for m in mag_limits[lunation][waveName]]))
    test_mags[0][0] = 5.

    skybrightness = [None]
    can_offset = [None]
    for flag, mag, sky, can_off in zip(flags_test, test_mags, skybrightness, can_offset):
        test_flags(flag, mag, mag_limits[lunation][waveName], lunation, waveName,
                   sky, offset_min_skybrightness, can_off)


    # test bright neighbor exclusion radius for very bright stars
    offset_min_skybrightness = 1
    waveName = 'Boss'
    lunation = 'bright'
    test_mags = numpy.array([0.1 if m != -999. else m for m in mag_limits[lunation][waveName]])
    r_exclude, _ = offset_definition(numpy.vstack((mag, mag)),
                                     mag_limits[lunation][waveName],
                                     lunation=lunation,
                                     waveName=waveName,
                                     fmagloss=fmagloss,
                                     obsSite='APO')
    assert numpy.all(r_exclude > 0.)

    offset_min_skybrightness = 1
    waveName = 'Apogee'
    lunation = 'bright'
    test_mags = numpy.array([0.1 if m != -999. else m for m in mag_limits[lunation][waveName]])
    r_exclude, _ = offset_definition(numpy.vstack((mag, mag)),
                                     mag_limits[lunation][waveName],
                                     lunation=lunation,
                                     waveName=waveName,
                                     fmagloss=fmagloss,
                                     obsSite='APO')
    assert numpy.all(r_exclude > 0.)



from coordio import calibration
import numpy
from coordio.conv import wokToPositioner, positionerToWok
import time
import matplotlib.pyplot as plt


def test_wok2positioner(plot=False):
    pt = calibration.positionerTable.reset_index()
    wc = calibration.wokCoords.reset_index()
    mt = pt.merge(wc, on=["site", "holeID"])

    nCoords = 10000
    alphas = numpy.random.uniform(size=nCoords)*360
    betas = numpy.random.uniform(size=nCoords)*178 + 2

    mtSamp = mt.sample(n=nCoords, replace=True)

    xBeta = mtSamp.bossX.to_numpy()
    yBeta = mtSamp.bossY.to_numpy()
    la = mtSamp.alphaArmLen.to_numpy()
    alphaOffDeg = mtSamp.alphaOffset.to_numpy()
    betaOffDeg = mtSamp.betaOffset.to_numpy()
    basePos = mtSamp[["xWok", "yWok", "zWok"]].to_numpy()
    iHat = mtSamp[["ix", "iy", "iz"]].to_numpy()
    jHat = mtSamp[["jx", "jy", "jz"]].to_numpy()
    kHat = mtSamp[["kx", "ky", "kz"]].to_numpy()
    dx = mtSamp.dx.to_numpy()
    dy = mtSamp.dy.to_numpy()

    tstart = time.time()
    xWok, yWok, zWok = positionerToWok(
        alphas, betas, xBeta, yBeta, la, alphaOffDeg, betaOffDeg,
        basePos, iHat, jHat, kHat, dx, dy
    )
    print("took", time.time()-tstart)
    # print(zWok)

    _alphas, _betas = wokToPositioner(
        xWok, yWok, zWok, xBeta, yBeta, la, alphaOffDeg, betaOffDeg,
        basePos, iHat, jHat, kHat, dx, dy
    )


    alphaErr = alphas - _alphas
    betaErr = betas - _betas


    numpy.testing.assert_allclose(alphas, _alphas)
    numpy.testing.assert_allclose(betas, _betas)

    print("std roundtrip", numpy.std(alphaErr), numpy.std(betaErr))

    if plot:
        plt.figure()
        plt.plot(xWok,yWok,'.k')
        plt.axis('equal')

        plt.figure()
        plt.plot(alphas, _alphas, '.k')

        plt.figure()
        plt.plot(betas, _betas, '.k')

        # plt.figure()
        # plt.hist(alphaErr, bins=numpy.linspace(-.01, .01, 100))

        # plt.figure()
        # plt.hist(betaErr, bins=numpy.linspace(-.01, .01, 100))

        plt.show()

    # import pdb; pdb.set_trace()


if __name__ == "__main__":
    test_wok2positioner(True)